# Copyright (C) 2011  Bradley N. Miller
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__author__ = "bmiller"

import os
from docutils import nodes
from runestone.common import RunestoneIdDirective, RunestoneIdNode


# try:
#     import conf
#     version = conf.version
#     staticserver = conf.staticserver
# except:
#     version = '2.1.0'
#     staticserver = 'runestonestatic.appspot.com'


def setup(app):
    app.add_directive("blockly", Blockly)

    app.add_node(BlocklyNode, html=(visit_block_html, depart_block_html))

    app.connect("doctree-resolved", process_activcode_nodes)
    app.connect("env-purge-doc", purge_activecodes)


#'
class BlocklyNode(nodes.General, nodes.Element, RunestoneIdNode):
    def __init__(self, content, **kwargs):
        """

        Arguments:
        - `self`:
        - `content`:
        """
        super(BlocklyNode, self).__init__(**kwargs)
        self.runestone_options = content


START = """
<html>
<head>
    <script src='blockly_compressed.js'> </script>
    <script src='blocks_compressed.js'> </script>
    <script src='javascript_compressed.js'> </script>
    <script src='python_compressed.js'> </script>
    <script src='msg/js/en.js'> </script>
    <link rel="stylesheet" href="bootstrap-3.4.1/css/bootstrap.min.css" type="text/css" />
    <link rel="stylesheet" href="video.css" type="text/css" />
    <script>
    // Get the objects we need to do logging from the parent frame.
    // This seems better than reloading all of jQuery and bookfuncs into the frame. But
    // Makes this a bit more dependent on the Runestone Environment.
    eBookConfig = parent.eBookConfig
    logBookEvent = parent.logBookEvent
    jQuery = parent.jQuery
    </script>
    <style>
      html, body {
        background-color: #fff;
        margin: 0;
        padding: 0;
      }
      .blocklySvg {
        height: 100%%;
        width: 100%%;
      }
      .active_out {
        margin-top: 5px;
        margin-left: 10px;
        margin-right: 5px;
      }
    </style>
</head>
<body>
<p>
    <button class="btn btn-default" onclick="showCode()">Show Python</button>
    <button class="btn btn-success" onclick="runCode()">Run</button>
</p>
<div id="%(divid)s" style="height: 480px; width: 600px;"></div>
"""

CTRL_START = """<xml id="toolbox" style="display: none">"""
CTRL_END = """</xml>"""


END = """
<script>
    Blockly.inject(document.getElementById('%(divid)s'),
        {path: './', toolbox: document.getElementById('toolbox')});

    function showCode() {
      // Generate JavaScript code and display it.
      Blockly.Python.INFINITE_LOOP_TRAP = null;
      var code = Blockly.Python.workspaceToCode();
      alert(code);
    }

    function runCode() {
      // Generate JavaScript code and run it.
      window.LoopTrap = 1000;
      Blockly.JavaScript.INFINITE_LOOP_TRAP = 'if (--window.LoopTrap == 0) throw "Infinite loop.";\\n';
      var code = Blockly.JavaScript.workspaceToCode();
      Blockly.JavaScript.INFINITE_LOOP_TRAP = null;
      if(logBookEvent) {
          logBookEvent({'event': 'blockly', 'act': 'run', 'div_id': '%(divid)s'});
      } else {
          console.log('logBookEvent is not defined.  This should be defined in the parent frame')
      }
      try {
        eval(code);
      } catch (e) {
        alert(e);
      }
    }

    Blockly.JavaScript['text_print'] = function(block) {
      // Print statement override.
      var argument0 = Blockly.JavaScript.valueToCode(block, 'TEXT',
          Blockly.JavaScript.ORDER_NONE) || '\\'\\'';
      return 'my_custom_print(' + argument0 + ', "%(divid)s" );\\n';
    };

    function my_custom_print(text,divid) {
      var p = document.getElementById(divid+"_pre");
      p.innerHTML += text + "\\n"
      }

    var xmlText = '%(preload)s';
    var xmlDom = Blockly.Xml.textToDom(xmlText);
    Blockly.Xml.domToWorkspace(Blockly.mainWorkspace, xmlDom);

  </script>

  <pre class="active_out" id="%(divid)s_pre"></pre>
  </body>
  </html>
"""
# self for these functions is an instance of the writer class.  For example
# in html, self is sphinx.writers.html.SmartyPantsHTMLTranslator
# The node that is passed as a parameter is an instance of our node class.


def visit_block_html(self, node):
    res = START % (node["runestone_options"])
    res += CTRL_START
    for ctrl in node["runestone_options"]["controls"]:
        if ctrl == "variables":
            res += '<category name="Variables" custom="VARIABLE"></category>'
        elif ctrl == "":
            pass
        elif ctrl[0] == "*":
            res += '<category name="%s">' % (ctrl[2:])
        elif ctrl == "====":
            res += "</category>"
        else:
            res += '<block type="%s"></block>\n' % (ctrl)
    res += CTRL_END
    res += END % (node["runestone_options"])
    path = os.path.join(
        node["runestone_options"]["blocklyHomePrefix"],
        "_static",
        node["runestone_options"]["divid"] + ".html",
    )
    final = (
        '<iframe class="blk-iframe" seamless src="%s" width="600" '
        'height="600"></iframe>' % path
    )
    f = open(path, "w")
    f.write(res)
    f.close()
    self.body.append(final)


def depart_block_html(self, node):
    """This is called at the start of processing an activecode node.  If activecode had recursive nodes
    etc and did not want to do all of the processing in visit_ac_html any finishing touches could be
    added here.
    """
    pass


def process_activcode_nodes(app, env, docname):
    pass


def purge_activecodes(app, env, docname):
    pass


class Blockly(RunestoneIdDirective):
    required_arguments = 1
    optional_arguments = 0
    has_content = True
    option_spec = RunestoneIdDirective.option_spec.copy()
    option_spec.update({})

    def run(self):
        super(Blockly, self).run()

        document = self.state.document
        rel_filename, filename = document.settings.env.relfn2path(self.arguments[0])

        pathDepth = rel_filename.count("/")
        self.options["blocklyHomePrefix"] = "../" * pathDepth

        plstart = len(self.content)
        if "preload::" in self.content:
            plstart = self.content.index("preload::")
            self.options["preload"] = " ".join(self.content[plstart + 1 :])

        if self.content:
            self.options["controls"] = self.content[:plstart]

        blockly_node = BlocklyNode(self.options, rawsource=self.block_text)
        blockly_node.source, blockly_node.line = self.state_machine.get_source_and_line(
            self.lineno
        )
        return [blockly_node]


"""
    Blockly.JavaScript['text_print'] = function(block) {
      // Print statement override.
      var argument0 = Blockly.JavaScript.valueToCode(block, 'TEXT',
          Blockly.JavaScript.ORDER_NONE) || '\'\'';
      return 'my_custom_print(' + argument0 + ');\n';
    };

    function my_custom_print(text) {
      var p = document.getElementById("blockly1_pre");
      p.innerHTML += text
      }

"""


# to preload blockly with a finished or partial program, do the following
#
# https://blockly-demo.appspot.com/static/apps/code/index.html?lang=en
#
# Now save the xml string.  And append something like the following to the script after blockly
# is created:
#
#       var xmlText = '<xml>  <block type="variables_set" id="1" inline="true" x="25" y="9">    <field name="VAR">X</field>    <value name="VALUE">      <block type="math_number" id="2">        <field name="NUM">10</field>      </block>    </value>  </block></xml>'
#       xmlDom = Blockly.Xml.textToDom(xmlText);
#       Blockly.Xml.domToWorkspace(Blockly.mainWorkspace, xmlDom);



# Copyright (C) 2011  Bradley N. Miller
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from docutils import nodes
from docutils.parsers.rst import directives
from runestone.common.runestonedirective import (
    RunestoneIdDirective,
    RunestoneIdNode,
)

__author__ = "bmiller"


def setup(app):
    app.add_directive("question", QuestionDirective)

    app.add_node(QuestionNode, html=(visit_question_html, depart_question_html))


class QuestionNode(nodes.General, nodes.Element, RunestoneIdNode):
    pass


def visit_question_html(self, node):
    # Set options and format templates accordingly
    env = node.document.settings.env
    if not hasattr(env, "questioncounter"):
        env.questioncounter = 0

    if "number" in node["runestone_options"]:
        env.questioncounter = int(node["runestone_options"]["number"])
    else:
        env.questioncounter += 1

    node["runestone_options"]["number"] = "start={}".format(env.questioncounter)

    res = TEMPLATE_START % node["runestone_options"]
    self.body.append(res)


def depart_question_html(self, node):
    # Set options and format templates accordingly
    res = TEMPLATE_END % node["runestone_options"]
    delimiter = "_start__{}_".format(node["runestone_options"]["divid"])

    self.body.append(res)


# Templates to be formatted by node options
TEMPLATE_START = """
    <div data-component="question" class="full-width container question %(optclass)s" data-question_label="%(question_label)s" id="%(divid)s" >
    <ol %(number)s class=arabic><li class="alert alert-warning">

    """
TEMPLATE_END = """
    </li></ol>
    </div>
    """


class QuestionDirective(RunestoneIdDirective):
    """
    .. question:: identifier
       :number: Force a number for this question

       Content  everything here is part of the question
       Content  It can be a lot...
    """

    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True
    has_content = True
    option_spec = RunestoneIdDirective.option_spec.copy()
    option_spec.update({"number": directives.positive_int})

    def run(self):
        super(QuestionDirective, self).run()
        self.assert_has_content()  # make sure question has something in it

        self.options["name"] = self.arguments[0].strip()

        question_node = QuestionNode()
        question_node["runestone_options"] = self.options
        (
            question_node["source"],
            question_node["line"],
        ) = self.state_machine.get_source_and_line(self.lineno)
        self.add_name(question_node)

        self.state.nested_parse(self.content, self.content_offset, question_node)

        return [question_node]



# Copyright (C) 2011  Bradley N. Miller
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__author__ = "bmiller"

from docutils import nodes
from docutils.parsers.rst import directives
from .pg_logger import exec_script_str_local
import json
import six
import time
import random
import requests
from runestone.server.componentdb import addQuestionToDB, addHTMLToDB
from runestone.common.runestonedirective import (
    RunestoneIdDirective,
    RunestoneIdNode,
)


def setup(app):
    app.add_directive("codelens", Codelens)

    app.add_config_value("codelens_div_class", "cd_section", "html")
    app.add_config_value("trace_url", "http://tracer.runestone.academy:5000", "html")
    app.add_node(
        CodeLensNode,
        html=(visit_codelens_html, depart_codelens_html),
        xml=(visit_codelens_xml, depart_codelens_xml),
    )


#  data-tracefile="pytutor-embed-demo/java.json"

VIS = """
<div class="runestone codelens %(optclass)s">
    <div class="%(divclass)s" data-component="codelens" data-question_label="%(question_label)s">
        <div id=%(divid)s_question class="ac_question">
        </div>
        <div class="pytutorVisualizer" id="%(divid)s"
           data-params='{"embeddedMode": true, "lang": "%(language)s", "jumpToEnd": false}'>
        </div>
        <p class="runestone_caption"><span class="runestone_caption_text">Activity: CodeLens %(caption)s (%(divid)s)</span> </p>
    </div>
"""


DATA = """
<script>
var %(divid)s_vis;

if (allTraceData === undefined) {
   var allTraceData = {};
}

allTraceData["%(divid)s"] = %(tracedata)s;
</script>
</div>
"""

PTX_TEMPLATE = """
<program xml:id="{divid}" interactive="codelens" language="{language}">
    <input>
{source}
    </input>
</program>
"""


class CodeLensNode(nodes.General, nodes.Element, RunestoneIdNode):
    pass


def visit_codelens_xml(self, node):
    html = VIS
    if "caption" not in node["runestone_options"]:
        node["runestone_options"]["caption"] = node["runestone_options"][
            "question_label"
        ]
    if "tracedata" in node["runestone_options"]:
        node["runestone_options"]["tracedata"] = (
            node["runestone_options"]["tracedata"]
            .replace("<", "&lt;")
            .replace(">", "&gt;")
        )

    res = PTX_TEMPLATE.format(**node["runestone_options"])
    self.output.append(res)


def visit_codelens_html(self, node):
    html = VIS
    if "caption" not in node["runestone_options"]:
        node["runestone_options"]["caption"] = node["runestone_options"][
            "question_label"
        ]
    if "tracedata" in node["runestone_options"]:
        html += DATA
    else:
        html += "</div>"
    html = html % node["runestone_options"]

    self.body.append(html)
    addHTMLToDB(
        node["runestone_options"]["divid"],
        node["runestone_options"]["basecourse"],
        html,
    )


def depart_codelens_html(self, node):
    pass


def depart_codelens_xml(self, node):
    pass


# Some documentation to help the author.
# Here's and example of a single stack frame.
# you might ask a qestion about the value of a global variable
# in which case the correct answer is expressed as:
#
# globals.a
#
# You could ask about a value on the heap
#
# heap.variable
#
# You could ask about a local variable -- not shown here.
#
# locals.variable
#
# You could even ask about what line is going to be executed next
#
# line
# {
#   "ordered_globals": [
#     "a",
#     "b"
#   ],
#   "stdout": "1\n",
#   "func_name": "<module>",
#   "stack_to_render": [],
#   "globals": {
#     "a": 1,
#     "b": 1
#   },
#   "heap": {},
#   "line": 5,
#   "event": "return"
# }


class Codelens(RunestoneIdDirective):
    """
    .. codelens:: uniqueid
       :tracedata: Autogenerated or provided
       :caption: caption below
       :showoutput: show stdout from program
       :question: Text of question to ask on breakline
       :correct: correct answer to the question
       :feedback: feedback for incorrect answers
       :breakline: Line to stop on and pop up a question dialog
       :python: either py2 or py3

        x = 0
        for i in range(10):
           x = x + i


    config values (conf.py):

    - codelens_div_class - custom CSS class of the component's outermost div
    """

    required_arguments = 1
    optional_arguments = 1
    option_spec = RunestoneIdDirective.option_spec.copy()
    option_spec.update(
        {
            "tracedata": directives.unchanged,
            "caption": directives.unchanged,
            "showoutput": directives.flag,
            "question": directives.unchanged,
            "correct": directives.unchanged,
            "feedback": directives.unchanged,
            "breakline": directives.nonnegative_int,
            "python": directives.unchanged,
            "language": directives.unchanged,
        }
    )

    has_content = True

    def run(self):
        super(Codelens, self).run()

        addQuestionToDB(self)

        self.JS_VARNAME = ""
        self.JS_VARVAL = ""

        def raw_dict(input_code, output_trace):
            ret = dict(code=input_code, trace=output_trace)
            return ret

        def js_var_finalizer(input_code, output_trace):
            ret = dict(code=input_code, trace=output_trace)
            json_output = json.dumps(ret, indent=None)
            return json_output

        if self.content:
            source = "\n".join(self.content)
        else:
            source = "\n"
        self.options["source"] = source.replace("<", "&lt;")
        CUMULATIVE_MODE = False
        self.JS_VARNAME = self.options["divid"] + "_trace"
        env = self.state.document.settings.env
        self.options["divclass"] = env.config.codelens_div_class

        if "showoutput" not in self.options:
            self.options["embedded"] = "true"  # to set embeddedmode to true
        else:
            self.options["embedded"] = "false"

        if "language" not in self.options:
            self.options["language"] = "python"

        if "python" not in self.options:
            if six.PY2:
                self.options["python"] = "py2"
            else:
                self.options["python"] = "py3"

        if "tracedata" not in self.options:
            if "question" in self.options:
                curTrace = exec_script_str_local(
                    source, None, CUMULATIVE_MODE, None, raw_dict
                )
                self.inject_questions(curTrace)
                json_output = json.dumps(curTrace, indent=None)
                self.options["tracedata"] = json_output
            else:
                if self.options["language"] == "python":
                    self.options["tracedata"] = exec_script_str_local(
                        source, None, CUMULATIVE_MODE, None, js_var_finalizer
                    )
                elif self.options["language"] == "java":
                    self.options["tracedata"] = self.get_trace(source, "java")
                elif self.options["language"] == "cpp":
                    self.options["tracedata"] = self.get_trace(source, "cpp")
                elif self.options["language"] == "c":
                    self.options["tracedata"] = self.get_trace(source, "c")
                else:
                    raise ValueError("language not supported")

        cl_node = CodeLensNode()
        cl_node["runestone_options"] = self.options
        cl_node["source"], cl_node["line"] = self.state_machine.get_source_and_line(
            self.lineno
        )
        return [cl_node]

    def inject_questions(self, curTrace):
        if "breakline" not in self.options:
            raise RuntimeError("Must have breakline option")
        breakline = self.options["breakline"]
        for frame in curTrace["trace"]:
            if frame["line"] == breakline:
                frame["question"] = dict(
                    text=self.options["question"],
                    correct=self.options["correct"],
                    div=self.options["divid"] + "_modal",
                    feedback=self.options["feedback"],
                )

    def get_trace(self, src, lang):
        env = self.state.document.settings.env
        url = f"{env.config.trace_url}/trace{lang}"

        tries = 0
        r = None
        while tries < 5:
            try:
                r = requests.post(url, data=dict(src=src), timeout=30)
            except requests.ReadTimeout:
                self.error(
                    "The request to the trace server timed out, you will need to rerun the build"
                )
                tries += 1
                time.sleep(random.random() * tries)
            except Exception as e:
                self.error(
                    "The request to the trace server failed, you will need to rerun the build "
                    + str(e)
                )
                tries += 1
                time.sleep(random.random() * tries)

            if r and r.status_code == 200:
                if lang == "java":
                    return r.text
                else:
                    res = r.text[r.text.find('{"code":') :]
                    return res
            else:
                tries += 1
                self.error(
                    f"Failed to get trace from server, status code: {r.status_code}"
                )
                time.sleep(random.random() * tries)
        if tries == 5:
            self.error("Failed to get trace from server after 5 tries")
            return ""
        return ""



"""
Configure logging for the application
=====================================

We keep the logging configuration quite simple as we want everything to go to stdout
so that the docker logs command will work.  We also want to be able to set the log level
from the environment variable ``LOG_LEVEL```.  The default is ``INFO``.
"""

#
# Imports
# =======
# These are listed in the order prescribed by `PEP 8`_.
#
# Standard library
# ----------------
import logging
import sys
import os


# write a function to set the log level based on the environment variable LOG_LEVEL
def get_log_level():
    log_level = os.getenv("LOG_LEVEL", "INFO")
    if log_level == "DEBUG":
        return logging.DEBUG
    elif log_level == "INFO":
        return logging.INFO
    elif log_level == "WARNING":
        return logging.WARNING
    elif log_level == "ERROR":
        return logging.ERROR
    elif log_level == "CRITICAL":
        return logging.CRITICAL
    else:
        return logging.INFO


# Third-party imports
# -------------------
# None.
#
# Local application imports
# -------------------------
# None.
#
#
# Logging
# =======
rslogger = logging.getLogger("runestone")
rslogger.setLevel(get_log_level())

handler = logging.StreamHandler(sys.stdout)
handler.setLevel(get_log_level())
formatter = logging.Formatter(
    "%(levelname)s - %(asctime)s - %(funcName)s - %(message)s"
)
handler.setFormatter(formatter)
rslogger.addHandler(handler)

# To turn on debugging for FastAPI and the database package:
#
# .. code:: Python
#   :number-lines:
#
#   from fastapi.logger import logger
#   import logging
#   logger.setLevel(logging.DEBUG)
#
# Otherwise, FastAPI logs at the default level.
# There is good info `here <https://medium.com/@PhilippeGirard5/fastapi-logging-f6237b84ea64>`__ on setting up logging.  There is also good info on hooking into the gunicorn logging system when you deploy `here  <https://pawamoy.github.io/posts/unify-logging-for-a-gunicorn-uvicorn-app/>`__.



# *********************************
# |docname| - Runestone Module init
# *********************************
import json

from .activecode import ActiveCode
from .animation import Animation
from .blockly import Blockly
from .codelens import Codelens
from .clickableArea import ClickableArea
from .datafile import DataFile
from .disqus import DisqusDirective
from .dragndrop import DragNDrop
from .fitb import FillInTheBlank
from .groupsub import GroupSubmission
from .hparsons import HParsonsDirective
from .khanex import Khanex
from .selectquestion import SelectQuestion
from .mchoice import MChoice, QuestionNumber
from .meta import Meta
from .parsons import ParsonsProblem
from .poll import Poll
from .quizly import Quizly
from .reveal import RevealDirective
from .shortanswer import JournalDirective
from .showeval import ShowEval
from .tabbedStuff import TabbedStuffDirective, TabDirective
from .timed import TimedDirective
from .video import Youtube, Vimeo, Video
from .webgldemo import WebglDemo


import os, socket
from importlib.metadata import version
import importlib.resources
import CodeChat.CodeToRest
from sphinx.errors import ExtensionError
from sphinx.builders.html import JavaScript


# TODO: clean up - many of the folders are not needed as the files are imported by webpack
#
# runestone_static_dirs()
# -----------------------
# Users can call this to get a list of all static directories in the runestone package
# normally this is just used by the `conf.py` file for building a runestone book
def runestone_static_dirs():
    basedir = os.path.dirname(__file__)
    module_static_js = [os.path.join(basedir, "dist")]
    module_static_js.append(os.path.join(basedir, "animation", "js"))
    module_static_js.append(os.path.join(basedir, "codelens", "js"))
    module_static_css = []
    module_static_css.append(os.path.join(basedir, "accessibility", "css"))
    module_static_css.append(os.path.join(basedir, "lp", "css"))
    return module_static_js + module_static_css + CodeChat.CodeToRest.html_static_path()


# runestone_extensions()
# -----------------------
# Users can call this to get a list of all extensions provided by runestone
# normally this is just used by the `conf.py` file for building a runestone book
deprecated_extensions = ["matrixeq", "webgldemo", "lp", "accessibility"]


def runestone_extensions():
    basedir = os.path.dirname(__file__)
    module_paths = [
        x
        for x in os.listdir(basedir)
        if os.path.isdir(os.path.join(basedir, x)) and x not in deprecated_extensions
    ]
    modules = [
        "runestone.{}".format(x)
        for x in module_paths
        if os.path.exists("{}/__init__.py".format(os.path.join(basedir, x)))
    ]
    # Place ``runestone.common`` first, so it can run init code needed by all other modules. This assumes that the first module in the list is run first. An alternative to this to guarantee this ordering is to call ``app.setup_extension('runestone.common')`` in every extension.
    modules.insert(0, modules.pop(modules.index("runestone.common")))
    return modules


# setup_js_defer(app, pagename, templatexname, context, doctree)
# -----------------------
# Used to inspect js right before it is rendered to page so that
# we can forcibly defer js files or prevent same
def setup_js_defer(app, pagename, templatexname, context, doctree):
    def js_defer(script_files):
        for js in sorted(script_files):
            if app.config.html_defer_js:
                # Files added from Runestone should already have defer set - so just add it to sphinx based ones
                to_defer = [
                    "_static/jquery.js",
                    "_static/underscore.js",
                    "_static/doctools.js",
                ]
                if isinstance(js, JavaScript) and js in to_defer:
                    js.attributes["defer"] = ""
            else:
                # config flag not set, prevent all deferrals
                if isinstance(js, JavaScript):
                    js.attributes.pop("defer", None)
        return ""

    context["js_defer"] = js_defer


# setup(app)
# ----------
def setup(app):
    """
    A normal Runestone project will import this function into its conf.py
    This setup will run after all of the extensions, so it is a good place
    for us to include our common javascript and css files.
    This could be expanded if there is additional initialization or customization
    we wanted to do for all projects.
    """
    # This tells the mathjax extension to always load mathjax.
    # This should save some configuration headaches as we transition from RST to PTX
    app.set_html_assets_policy("always")
    # Include JS and CSS produced by webpack. See `webpack static imports <webpack_static_imports>`_.
    with importlib.resources.path("runestone.dist", "webpack_static_imports.json") as p:
        with open(p) as f:
            wb_imports = json.load(f)
            script_files = wb_imports["js"]
            _css_files = css_files + wb_imports["css"]

    for jsfile in script_files:
        try:
            app.add_autoversioned_javascript(jsfile, defer="")
        except ExtensionError:
            app.add_js_file(jsfile, defer="")
    for cssfile in _css_files:
        try:
            app.add_autoversioned_stylesheet(cssfile)
        except ExtensionError:
            app.add_css_file(cssfile)

    # projects can define their own custom css or js files to include
    # But since this is imported into their conf.py authors must
    # define them as attributes of the setup function
    # ``setup.custom_xxx_files``
    try:
        for c in setup.custom_css_files:
            app.add_css_file(c)
        print("Adding custom CSS files")
    except AttributeError:
        print("No custom CSS files")
    try:
        for c in setup.custom_js_files:
            if isinstance(c, dict):
                # peel off filename, pass rest of key/values on as kwargs
                filename = c.pop("file")
                app.add_autoversioned_javascript(filename, **c)
            else:
                app.add_autoversioned_javascript(c)
        print("Adding custom Javascript")
    except AttributeError:
        print("No custom js files")

    app.config.html_static_path.append("dist/")
    app.add_config_value("html_defer_js", False, "env")
    app.connect("html-page-context", setup_js_defer)


def get_master_url():
    if socket.gethostname() in ["runestone-deploy", "rsbuilder"]:
        master_url = "https://runestone.academy"
    elif "RUNESTONE_HOST" in os.environ:
        port = os.environ.get("RUNESTONE_PORT", 80)
        secure = os.environ.get("RUNESTONE_PROTOCOL", "http")
        master_url = "{}://{}:{}".format(secure, os.environ["RUNESTONE_HOST"], port)
    else:
        master_url = "http://127.0.0.1:8000"

    return master_url


from paver.easy import task, cmdopts, sh
from sphinxcontrib import paverutils


@task
@cmdopts(
    [
        ("all", "a", "rebuild everything"),
        ("outputdir=", "o", "output static files here"),
        ("masterurl=", "u", "override the default master url"),
        ("masterapp=", "p", "override the default master app"),
    ]
)
def build(options):
    if "all" in options.build:
        options["force_all"] = True
        options["freshenv"] = True

    try:
        bi = sh("git describe --long", capture=True)[:-1]
        bnum = bi.split("-")[0]
        options.build.template_args["build_info"] = bnum
    except:
        bi = "unknown-0-0"
        options.build.template_args["build_info"] = "unknown"

    with open("build_info", "w") as bif:
        bif.write(bi)
        bif.write("\n")

    if "outputdir" in options.build:
        options.build.outdir = options.build.outputdir

    if "masterurl" in options.build:
        options.build.template_args["course_url"] = options.build.masterurl

    if "masterapp" in options.build:
        options.build.template_args["appname"] = options.build.masterapp

    print("Building into ", options.build.outdir)
    rc = paverutils.run_sphinx(options, "build")

    if rc == 0 or rc is None:
        print("Done, build successful")
    else:
        print("Error in building code {}".format(rc))

    return rc


# Module variables
# ----------------
try:
    runestone_version = version("runestone")
except:
    print("Could not determine runestone version")
    runestone_version = "unknown"


css_files = [
    # Generated from a template, so it can't be directly included in the webpack.
    "bootstrap-sphinx.css",
]


cmap = {
    "activecode": ActiveCode,
    "mchoice": MChoice,
    "fillintheblank": FillInTheBlank,
    "timed": TimedDirective,
    "qnum": QuestionNumber,
    "codelens": Codelens,
    "clickablearea": ClickableArea,
    "datafile": DataFile,
    "disqus": DisqusDirective,
    "dragndrop": DragNDrop,
    "groupsub": GroupSubmission,
    "hparsons": HParsonsDirective,
    "parsonsprob": ParsonsProblem,
    "poll": Poll,
    "quizly": Quizly,
    "reveal": RevealDirective,
    "selectquestion": SelectQuestion,
    "shortanswer": JournalDirective,
    "showeval": ShowEval,
    "tabbed": TabbedStuffDirective,
    "tab": TabDirective,
    "video": Video,
    "youtube": Youtube,
    "vimeo": Vimeo,
}



# .. Copyright (C) 2021 Bryan A. Jones.
#
# *********************************************************
# |docname| - A Runestone extension supporting cellbotics
# *********************************************************
# This extension provides one new directive, ``ble-pair-button``, to add the Bluetooth Low Energy "Pair" button and status display to a page where users will be using the Cellbotics Python module and associated BLE functionality. The extension also includes all the supporting JavaScript files needed for the Cellbotics Python module.
#
#
# Imports
# =======
# These are listed in the order prescribed by `PEP 8
# <http://www.python.org/dev/peps/pep-0008/#imports>`_.
#
# Standard library
# ----------------
# None.
#
# Third-party imports
# -------------------
from docutils import nodes
from docutils.parsers.rst import Directive

# Local imports
# -------------
# None


# Code
# ====
class BlePairNode(nodes.General, nodes.Element):
    pass


def visit_ble_pair_html(self, node):
    self.body.append(
        '<div data-component="ble">\n'
        '   <script>runestone_import("ble");</script>\n'
        '   <button id="ble_pair_button" type="button" disabled>Pair</button>\n'
        '   <span id="ble_pair_status"></span>\n'
        '</div>\n'
    )


def depart_ble_pair_html(self, node):
    pass


class BlePairDirective(Directive):
    has_content = False
    required_arguments = 0
    optional_arguments = 0
    option_spec = {}

    def run(self):
        return [BlePairNode()]


def setup(app):
    # Add the Pair button directive.
    app.add_node(BlePairNode, html=(visit_ble_pair_html, depart_ble_pair_html))
    app.add_directive('ble-pair-button', BlePairDirective)



from rsptx.data_types import autograde
from rsptx.data_types import which_to_grade
from rsptx.data_types import question_type
from rsptx.data_types import language

__all__ = ["autograde", "which_to_grade", "question_type", "language"]



# Copyright (C) 2011  Bradley N. Miller
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# This files runs during runestone build. It generates the node data
# that is used to render the component by the  quizly.js script.

# Note: An import entry for quizly must be included in runestone/__init__.py

# Note: The content files for the khanex component must be stored in the
# course's _static folder. Download the following file and unzip it in _static:
# https://github.com/ram8647/khanex/blob/16398fd496fad5b93fdf4d72c274064db8d1d1ac/khanex-runestone.zip

import shutil
import os
from pathlib import Path
from runestone.server.componentdb import addQuestionToDB, addHTMLToDB
from runestone.common import RunestoneIdDirective, RunestoneIdNode
from docutils import nodes

__author__ = "rmorelli"

# Debug flags
DEBUG = False
VERBOSE = False


# Template that will load index.html?quizname=the-quiz-name into an <iframe>
# NOTE: Hardcoding the container class.  Temporarily??
KHANEX_TEMPLATE = """
    <div class="runestone alert alert-warning %(optclass)s">
     <div  data-component="khanex" id="%(divid)s" data-question_label="%(question_label)s" style="visibility: visible;">
       <iframe style="transform:scale(0.7,1);transform-origin:0 0;" width="990px" height="490px" src="../_static/khanex/qs/###.html"></iframe>
     </div>
    </div>
       """

# Define the khanex directive


def setup(app):
    app.add_directive("khanex", Khanex)
    app.add_node(KhanexNode, html=(visit_khanex_html, depart_khanex_html))


# The only content needed from khanex.py is the exercise name


class KhanexNode(nodes.General, nodes.Element, RunestoneIdNode):
    pass


# self for these functions is an instance of the writer class.  For example
# in html, self is sphinx.writers.html.SmartyPantsHTMLTranslator
# The node that is passed as a parameter is an instance of our node class.


def visit_khanex_html(self, node):

    node["delimiter"] = "_start__{}_".format(node["runestone_options"]["divid"])
    self.body.append(node["delimiter"])

    print("DEBUG: visit_khanex_html exername = " + node["exername"]) if DEBUG else None
    print("DEBUG: visit_khanex_html template = " + node["template"]) if DEBUG else None
    print(
        "DEBUG: visit_khanex_html options = " + str(node["runestone_options"])
    ) if DEBUG else None

    res = node["template"] % (node["runestone_options"])
    print("DEBUG: visit_khanex_html res = " + res) if DEBUG else None
    self.body.append(res)


def depart_khanex_html(self, node):
    """This is called at the start of processing an activecode node.  If activecode had recursive nodes
    etc and did not want to do all of the processing in visit_ac_html any finishing touches could be
    added here.
    """
    print("DEBUG: depart_khanex_html") if DEBUG else None
    bc = node["runestone_options"]["basecourse"]
    addHTMLToDB(
        node["runestone_options"]["divid"],
        bc,
        "".join(self.body[self.body.index(node["delimiter"]) + 1 :]).replace(
            "../_static", f"/ns/books/published/{bc}/_static"
        ),
    )
    self.body.remove(node["delimiter"])
    pass


def process_activcode_nodes(app, env, docname):
    pass


def purge_activecodes(app, env, docname):
    pass


# The object created when the khanex directive is parsed in an RST.
# The exername is the only required argument.
class Khanex(RunestoneIdDirective):
    """
    .. khanex:: some_unique_id, e.g., ex1

       :exercise: unique-ex-name:
    """

    required_arguments = 1
    optional_arguments = 0
    has_content = True
    option_spec = RunestoneIdDirective.option_spec.copy()
    option_spec.update({})

    # This invokes the rendering code in js/khanex.js
    def run(self):
        super(Khanex, self).run()
        print("DEBUG; Khanex.run()") if DEBUG else None

        addQuestionToDB(self)  # Not sure whether this works?
        document = self.state.document
        rel_filename, filename = document.settings.env.relfn2path(self.arguments[0])

        pathDepth = rel_filename.count("/")
        self.options["khanexHomePrefix"] = "./" * pathDepth

        plstart = len(self.content)
        if self.content:
            self.options["controls"] = self.content[:plstart]

        khanex_node = KhanexNode()
        khanex_node["runestone_options"] = self.options
        khanex_node["exername"] = str(self.options["controls"][0])
        khanex_node["exername"] = str.strip(khanex_node["exername"][10:])
        khanex_node["template"] = KHANEX_TEMPLATE.replace(
            "###", khanex_node["exername"]
        )

        (
            khanex_node["source"],
            khanex_node["line"],
        ) = self.state_machine.get_source_and_line(self.lineno)
        return [khanex_node]



# Copyright (C) 2011  Bradley N. Miller
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# **************************************************************
# |docname| - Record chapter info in the database during a build
# **************************************************************

__author__ = "bmiller"

from typing import Sequence
from collections import OrderedDict
import docutils
from sphinx.util import logging

logger = logging.getLogger(__name__)
logger.setLevel(30)

ignored_chapters = ["", "FrontBackMatter", "Appendices"]


def setup(app):
    """
    The setup function ensures that we install this module as a Sphinx extension. Even though
    we are not going for a new directive or role we can use the extension mechanism to add
    specific event handlers.
    """
    global dburl, engine, meta, sess, questions, assignments, assignment_questions, courses, competency, chapters, sub_chapters, cname
    # app.connect("env-get-updated", test_updated)
    app.connect("doctree-resolved", env_updated)


def env_updated(app, doctree, docname):
    """
    This may be the best place to walk the completed document with TOC
    """
    # ``docname`` is stored with Unix-style forward slashes, even on Windows. Therefore, we can't use ``os.path.basename`` or ``os.sep``.
    splits = docname.split("/")
    # If the docname is ``'index'``, then set ``chap_id`` to an empty string.
    chap_id = splits[-2] if len(splits) > 1 else ""
    subchap_id = splits[-1]

    if not hasattr(app.env, "chap_titles"):
        app.env.chap_titles = {}
        app.env.chap_numbers = {}
        app.env.subchap_titles = {}
        app.env.subchap_numbers = {}
        app.env.skips = {}

    if "toctree" in docname:
        secnum_tuple = app.env.toc_secnumbers.get(docname, {}).get("")
        chap_num = secnum_tuple[0] if secnum_tuple else 999
        app.env.chap_numbers[chap_id] = chap_num
        chap_title = (
            list(doctree.traverse(docutils.nodes.section))[0]
            .next_node(docutils.nodes.Titular)
            .astext()
        )
        app.env.chap_titles[chap_id] = f"{chap_num}. {chap_title}"
        return []

    for section in doctree.traverse(docutils.nodes.section):
        # Find the section number of the current document. See `../common/question_number.py` for details.
        secnum_tuple = app.env.toc_secnumbers.get(docname, {}).get("")
        # Gievn a section number as a tuple, such as ``(1, 2, 3)``, turn this into the string "1.2.3 ".
        secnum_str = ".".join(map(str, secnum_tuple)) + " " if secnum_tuple else ""
        # Prepend it to the title.
        title = secnum_str + section.next_node(docutils.nodes.Titular).astext()

        if hasattr(app.env, "skipreading") and docname in app.env.skipreading:
            app.env.skips[(chap_id, subchap_id)] = True

        if chap_id in ignored_chapters or subchap_id == "index":
            continue
        if chap_id not in app.env.chap_titles:
            chap_num = secnum_tuple[0] if secnum_tuple else 999
            # Make sure we have at least the default chapter title for this chapter
            # otherwise it will be set properly in the toctree processing above.
            if not app.env.chap_titles.get(chap_id, False):
                app.env.chap_titles[chap_id] = f"{chap_num}. {chap_id}"
            app.env.chap_numbers[chap_id] = chap_num

        if chap_id not in app.env.subchap_titles:
            app.env.subchap_titles[chap_id] = OrderedDict()
            app.env.subchap_numbers[chap_id] = OrderedDict()
        if subchap_id not in app.env.subchap_titles[chap_id]:
            app.env.subchap_titles[chap_id][subchap_id] = title
            app.env.subchap_numbers[chap_id][subchap_id] = make_subchap_num(
                secnum_tuple
            )
            logger.debug(
                f"{chap_id} - {subchap_id} {app.env.subchap_numbers[chap_id][subchap_id]}    {secnum_tuple}                "
            )

    return []


# Given a sequence of section numbers, produce a single (orderable) int from the subchapter number and optionally the subsubchapter. Any finer divisions are ignored.
def make_subchap_num(
    # A sequence of section numbers in the format ``(chapter, subchapter, optional_subsubchapter, ignored_stuff...)``.
    sn_seq: Sequence[int],
    # Returns subchapter*100 + optional_subsubchapter.
) -> int:
    if not sn_seq or len(sn_seq) < 2:
        return 0

    # Ignore the chapter number at ``sn_tuple[0]`` -- we're creating a subchapter number.
    ret = sn_seq[1] * 100
    # Include the subsubchapter number if it exists. Ignore everything else.
    if len(sn_seq) > 2:
        ret += sn_seq[2]
    return ret



# Copyright (C) 2011  Bradley N. Miller
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__author__ = "isaiahmayerchak"
# acbart did most of this code, I mostly just changed the template

from docutils import nodes
from docutils.parsers.rst import directives
from runestone.mchoice import Assessment
from runestone.server.componentdb import (
    addQuestionToDB,
    addHTMLToDB,
    maybeAddToAssignment,
)
from runestone.common.runestonedirective import (
    RunestoneDirective,
    RunestoneIdNode,
)


def setup(app):
    app.add_directive("shortanswer", JournalDirective)
    app.add_node(
        JournalNode,
        html=(visit_journal_html, depart_journal_html),
        xml=(visit_journal_xml, depart_journal_xml),
    )

    app.add_config_value("shortanswer_div_class", "journal", "html")
    app.add_config_value(
        "shortanswer_optional_div_class", "journal alert alert-success", "html"
    )


TEXT_START = """
<div class="runestone">
<div data-component="shortanswer" data-question_label="%(question_label)s" class="%(divclass)s" id=%(divid)s %(optional)s %(mathjax)s %(attachment)s >
"""

TEXT_END = """
</div>
</div> <!-- end of runestone div -->
"""

XML_START = """
<exercise label="{divid}"  {optional}>
    <statement>
"""

XML_END = """
    </statement>
</exercise>
"""


class JournalNode(nodes.General, nodes.Element, RunestoneIdNode):
    pass


def visit_journal_html(self, node):

    node["delimiter"] = "_start__{}_".format(node["runestone_options"]["divid"])
    self.body.append(node["delimiter"])

    res = TEXT_START % node["runestone_options"]

    self.body.append(res)


def visit_journal_xml(self, node):

    res = XML_START.format(**node["runestone_options"])

    self.output.append(res)


def depart_journal_html(self, node):

    components = dict(node["runestone_options"])

    res = TEXT_END % components
    self.body.append(res)

    addHTMLToDB(
        node["runestone_options"]["divid"],
        components["basecourse"],
        "".join(self.body[self.body.index(node["delimiter"]) + 1 :]),
    )

    self.body.remove(node["delimiter"])


def depart_journal_xml(self, node):
    self.output.append(XML_END)


class JournalDirective(Assessment):
    """
    .. shortanswer:: uniqueid
       :optional:

       text of the question goes here


    config values (conf.py):

    - shortanswer_div_class - custom CSS class of the component's outermost div
    """

    required_arguments = 1  # the div id
    optional_arguments = 0
    final_argument_whitespace = True
    has_content = True
    option_spec = Assessment.option_spec.copy()
    option_spec.update({"mathjax": directives.flag, "attachment": directives.flag})

    node_class = JournalNode

    def run(self):
        super(JournalDirective, self).run()
        addQuestionToDB(self)
        # Raise an error if the directive does not have contents.
        self.assert_has_content()

        self.options["mathjax"] = "data-mathjax" if "mathjax" in self.options else ""
        self.options["attachment"] = (
            "data-attachment" if "attachment" in self.options else ""
        )

        journal_node = JournalNode()
        journal_node["runestone_options"] = self.options
        (
            journal_node["source"],
            journal_node["line"],
        ) = self.state_machine.get_source_and_line(self.lineno)

        self.updateContent()

        self.state.nested_parse(self.content, self.content_offset, journal_node)

        env = self.state.document.settings.env
        if self.options["optional"]:
            self.options["divclass"] = env.config.shortanswer_optional_div_class
        else:
            self.options["divclass"] = env.config.shortanswer_div_class

        maybeAddToAssignment(self)

        return [journal_node]



# Copyright (C) 2011  Bradley N. Miller
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__author__ = "isaacdontjelindell"

from docutils import nodes
from docutils.parsers.rst import directives
from runestone.common.runestonedirective import (
    RunestoneDirective,
    RunestoneNode,
)


DISQUS_BOX = """\
<script>
    function %(identifier)s_disqus(source) {
        if (window.DISQUS) {

            $('#disqus_thread').insertAfter(source); //put the DIV for the Disqus box after the link

            //if Disqus exists, call it's reset method with new parameters
            DISQUS.reset({
                reload: true,
                config: function () {
                    this.page.identifier = '%(identifier)s_' + eBookConfig.course;
                    this.page.url = 'https://%(identifier)s_'+eBookConfig.course+'.interactivepython.com/#!';
                }
            });

        } else {
            //insert a wrapper in HTML after the relevant "show comments" link
            $('<div id="disqus_thread"></div>').insertAfter(source);

            // set Disqus required vars
            disqus_shortname = '%(shortname)s';
            disqus_identifier = '%(identifier)s_' + eBookConfig.course;
            disqus_url = 'http://%(identifier)s_'+eBookConfig.course+'.interactivepython.com/#!';

            //append the Disqus embed script to HTML
            var dsq = document.createElement('script'); dsq.type = 'text/javascript'; dsq.async = true;
            dsq.src = 'https://' + disqus_shortname + '.disqus.com/embed.js';
            $('head').append(dsq);

        }
    }
</script>
"""

DISQUS_LINK = """
<a href="#disqus_thread" class="disqus_thread_link" data-disqus-identifier="%(identifier)s" onclick="%(identifier)s_disqus(this);">Show Comments</a>
<script>
  window.addEventListener('load', (event) => {
    $("a[data-disqus-identifier='%(identifier)s']").attr('data-disqus-identifier', '%(identifier)s_' + eBookConfig.course);
  });
</script>
"""


def setup(app):
    app.add_directive("disqus", DisqusDirective)

    app.add_node(DisqusNode, html=(visit_disqus_html, depart_disqus_html))
    app.connect("doctree-resolved", process_disqus_nodes)
    app.connect("env-purge-doc", purge_disqus_nodes)


class DisqusNode(nodes.General, nodes.Element, RunestoneNode):
    pass


def visit_disqus_html(self, node):
    res = DISQUS_BOX
    res += DISQUS_LINK

    res = res % node["runestone_options"]

    self.body.append(res)


def depart_disqus_html(self, node):
    pass


def process_disqus_nodes(app, env, docname):
    pass


def purge_disqus_nodes(app, env, docname):
    pass


class DisqusDirective(RunestoneDirective):
    """
    .. disqus::
       :shortname: Your registered disqus id
       :identifier: unique id for this discussion
    """

    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = True
    has_content = False
    option_spec = RunestoneDirective.option_spec.copy()
    option_spec.update(
        {
            "shortname": directives.unchanged_required,
            "identifier": directives.unchanged_required,
        }
    )

    def run(self):
        """
        generate html to include Disqus box.
        :param self:
        :return:
        """

        disqus_node = DisqusNode()
        disqus_node["runestone_options"] = self.options
        (
            disqus_node["source"],
            disqus_node["line"],
        ) = self.state_machine.get_source_and_line(self.lineno)
        return [disqus_node]



# Copyright (C) 2011  Bradley N. Miller
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__author__ = "bmiller"

from runestone.common.runestonedirective import RunestoneIdDirective

_base_js_escapes = (
    ("\\", r"\u005C"),
    ("'", r"\u0027"),
    ('"', r"\u0022"),
    ("'", r"\u0027"),
    (">", r"\u003E"),
    ("<", r"\u003C"),
    ("&", r"\u0026"),
    ("=", r"\u003D"),
    ("-", r"\u002D"),
    (";", r"\u003B"),
    ("\u2028", r"\u2028"),
    ("\u2029", r"\u2029"),
)

# Escape every ASCII character with a value less than 32.
_js_escapes = _base_js_escapes + tuple([("%c" % z, "\\u%04X" % z) for z in range(32)])

# escapejs from Django: https://www.djangoproject.com/
def escapejs(value):
    """Hex encodes characters for use in JavaScript strings."""
    if not isinstance(value, str):
        value = str(value)

    for bad, good in _js_escapes:
        value = value.replace(bad, good)

    return value


class Assessment(RunestoneIdDirective):
    """Base Class for assessments"""

    def run(self):
        super(Assessment, self).run()

        if self.content:
            if "iscode" in self.options:
                self.options["bodytext"] = "<pre>" + "\n".join(self.content) + "</pre>"
            else:
                self.options["bodytext"] = "\n".join(self.content)
        else:
            self.options["bodytext"] = "\n"



# *********
# |docname|
# *********
# Copyright (C) 2011  Bradley N. Miller
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import print_function

__author__ = "bmiller"

from docutils import nodes
from docutils.parsers.rst import directives
from .textfield import textfield_role
from sqlalchemy import Table
from runestone.server.componentdb import (
    addQuestionToDB,
    addHTMLToDB,
    get_engine_meta,
    maybeAddToAssignment,
)
from runestone.common.runestonedirective import (
    RunestoneIdDirective,
    RunestoneIdNode,
)

try:
    from html import escape  # py3
except ImportError:
    from cgi import escape  # py2


def setup(app):
    app.add_directive("activecode", ActiveCode)
    app.add_directive("actex", ActiveExercise)
    app.add_role("textfield", textfield_role)
    app.add_config_value(
        "activecode_div_class",
        "runestone explainer ac_section",
        "html",
    )
    app.add_config_value("activecode_hide_load_history", False, "html")
    app.add_config_value("wasm_uri", "/_static", "html")

    app.add_node(
        ActivecodeNode,
        html=(visit_ac_html, depart_ac_html),
        xml=(visit_ac_xml, depart_ac_xml),
    )

    app.connect("doctree-resolved", process_activcode_nodes)
    app.connect("env-purge-doc", purge_activecodes)


XML_EX_START = """
<exercise label="{divid}">
    <statement>
"""

XML_EX_END = """
    </statement>
    <program xml:id="{divid}_editor" interactive='activecode' language="{language}">
{optional_prefix_code}
        <input>
{xml_source_code}
        </input>
{optional_suffix_code}
    </program>
</exercise>
"""

XML_LISTING_START = """
    <program xml:id="{divid}" interactive='activecode' language="{language}">
{optional_prefix_code}
        <input>
{xml_source_code}
        </input>
{optional_suffix_code}
    </program>
"""


TEMPLATE_START = """
<div class="%(divclass)s %(optclass)s">
<div data-component="activecode" id=%(divid)s data-question_label="%(question_label)s">
<div id=%(divid)s_question class="ac_question">
"""

TEMPLATE_END = """
</div>
<textarea data-lang="%(language)s" id="%(divid)s_editor" %(autorun)s
    %(hidecode)s %(include)s %(timelimit)s %(coach)s %(codelens)s %(enabledownload)s %(chatcodes)s %(optional)s
    data-audio='%(ctext)s' %(sourcefile)s %(datafile)s %(stdin)s %(tie)s %(dburl)s %(nopair)s
    %(cargs)s %(largs)s %(rargs)s %(iargs)s %(gradebutton)s %(caption)s %(hidehistory)s %(wasmuri)s
    %(showlastsql)s style="visibility: hidden;">
%(initialcode)s
</textarea>
</div>
</div>
"""


class ActivecodeNode(nodes.General, nodes.Element, RunestoneIdNode):
    pass


def visit_ac_xml(self, node):
    node["runestone_options"]["xml_source_code"] = (
        node["runestone_options"]["xml_source_code"]
        .replace("<", "&lt;")
        .replace(">", "&gt;")
    )
    if node["runestone_options"]["has_problem_statement"]:
        res = XML_EX_START.format(**node["runestone_options"])
        self.output.append(res)


def depart_ac_xml(self, node):
    if node["runestone_options"]["xml_prefix_code"]:
        node["runestone_options"]["optional_prefix_code"] = (
            "<prefix>\n"
            + node["runestone_options"]["xml_prefix_code"]
            .replace("<", "&lt;")
            .replace(">", "&gt;")
            + "\n</prefix>\n"
        )
    else:
        node["runestone_options"]["optional_prefix_code"] = ""
    if node["runestone_options"]["xml_suffix_code"]:
        node["runestone_options"]["optional_suffix_code"] = (
            "<tests>\n"
            + node["runestone_options"]["xml_suffix_code"]
            .replace("<", "&lt;")
            .replace(">", "&gt;")
            + "\n</tests>\n"
        )
    else:
        node["runestone_options"]["optional_suffix_code"] = ""
    if node["runestone_options"]["has_problem_statement"]:
        res = XML_EX_END.format(**node["runestone_options"])
        self.output.append(res)
    else:
        res = XML_LISTING_START.format(**node["runestone_options"])
        self.output.append(res)


# self for these functions is an instance of the writer class.  For example
# in html, self is sphinx.writers.html.SmartyPantsHTMLTranslator
# The node that is passed as a parameter is an instance of our node class.


def visit_ac_html(self, node):
    # print self.settings.env.activecodecounter

    node["delimiter"] = "_start__{}_".format(node["runestone_options"]["divid"])

    self.body.append(node["delimiter"])

    res = TEMPLATE_START % node["runestone_options"]
    self.body.append(res)


def depart_ac_html(self, node):
    """This is called at the start of processing an activecode node.  If activecode had recursive nodes
    etc and did not want to do all of the processing in visit_ac_html any finishing touches could be
    added here.
    """
    res = TEMPLATE_END % node["runestone_options"]
    self.body.append(res)

    addHTMLToDB(
        node["runestone_options"]["divid"],
        node["runestone_options"]["basecourse"],
        "".join(self.body[self.body.index(node["delimiter"]) + 1 :]),
    )

    self.body.remove(node["delimiter"])


def process_activcode_nodes(app, env, docname):
    pass


def purge_activecodes(app, env, docname):
    pass


class ActiveCode(RunestoneIdDirective):
    """
    .. activecode:: uniqueid
       :nocanvas:  -- do not create a canvas
       :autograde: unittest
       :nopre: -- do not create an output component
       :above: -- put the canvas above the code
       :autorun: -- run this activecode as soon as the page is loaded
       :caption: this is the caption
       :include: div1,div2 -- invisibly include code from another activecode
       :hidecode: -- Don't show the editor initially
       :nocodelens: -- Do not show the codelens button
       :timelimit: -- set the time limit for this program in seconds
       :language: python, html, javascript, java, python2, python3
       :chatcodes: -- Enable users to talk about this code snippet with others
       :tour_1: audio tour track
       :tour_2: audio tour track
       :tour_3: audio tour track
       :tour_4: audio tour track
       :tour_5: audio tour track
       :stdin: : A file to simulate stdin (java, python2, python3)
       :datafile: : A datafile for the program to read (java, python2, python3)
       :sourcefile: : source files (java, python2, python3)
       :available_files: : other additional files (java, python2, python3)
       :enabledownload: -- allow textfield contents to be downloaded as *.py file
       :nopair: -- disable pair programming features
       :dburl: url to load database for sql mode
       :showlastsql: -- Only show the last sql result in output

        If this is a homework problem instead of an example in the text
        then the assignment text should go here.  The assignment text ends with
        the line containing four tilde ~
        ~~~~
        print("Hidden code before students code - good for scaffolding")
        ^^^^
        print("hello world")
        ====
        print("Hidden code, such as unit tests come after the four = signs")

    config values (conf.py):

    - activecode_div_class - custom CSS class of the component's outermost div
    - activecode_hide_load_history - if True, hide the load history button
    - wasm_uri - Path or Full URL to folder containing WASM files for SQL. /_static is default
    """

    required_arguments = 1
    optional_arguments = 1
    has_content = True
    option_spec = RunestoneIdDirective.option_spec.copy()
    option_spec.update(
        {
            "nocanvas": directives.flag,
            "nopre": directives.flag,
            "above": directives.flag,  # put the canvas above the code
            "autorun": directives.flag,
            "caption": directives.unchanged,
            "include": directives.unchanged,
            "hidecode": directives.flag,
            "language": directives.unchanged,
            "chatcodes": directives.flag,
            "tour_1": directives.unchanged,
            "tour_2": directives.unchanged,
            "tour_3": directives.unchanged,
            "tour_4": directives.unchanged,
            "tour_5": directives.unchanged,
            "nocodelens": directives.flag,
            "coach": directives.flag,
            "gradebutton": directives.flag,
            "timelimit": directives.unchanged,
            "stdin": directives.unchanged,
            "datafile": directives.unchanged,
            "sourcefile": directives.unchanged,
            "available_files": directives.unchanged,
            "enabledownload": directives.flag,
            "compileargs": directives.unchanged,
            "linkargs": directives.unchanged,
            "interpreterargs": directives.unchanged,
            "runargs": directives.unchanged,
            "tie": directives.unchanged,
            "nopair": directives.flag,
            "dburl": directives.unchanged,
            "showlastsql": directives.flag,
        }
    )

    def run(self):
        super(ActiveCode, self).run()

        env = self.state.document.settings.env
        # keep track of how many activecodes we have....
        # could be used to automatically make a unique id for them.
        if not hasattr(env, "activecodecounter"):
            env.activecodecounter = 0
        env.activecodecounter += 1
        self.options["name"] = self.arguments[0].strip()

        explain_text = None
        if self.content:
            if "~~~~" in self.content:
                idx = self.content.index("~~~~")
                explain_text = self.content[:idx]
                self.content = self.content[idx + 1 :]
            source = "\n".join(self.content)
        else:
            source = "\n"

        if explain_text:
            self.options["has_problem_statement"] = True
            self.explain_text = explain_text
        else:
            self.options["has_problem_statement"] = False
            self.explain_text = ["Not an Exercise"]

        if self.content and "unittest.gui" in source:
            self.options["autograde"] = "unittest"

        addQuestionToDB(self)

        self.options["initialcode"] = source
        str = source.replace("\n", "*nline*")
        str0 = str.replace('"', "*doubleq*")
        str1 = str0.replace("(", "*open*")
        str2 = str1.replace(")", "*close*")
        str3 = str2.replace("'", "*singleq*")
        self.options["argu"] = str3

        # TODO: This is BAD -- using '_' as a key for audio tour stuff is wrong.
        complete = ""
        no_of_buttons = 0
        okeys = list(self.options.keys())
        for k in okeys:
            if "tour_" in k:
                x, label = k.split("_")
                no_of_buttons = no_of_buttons + 1
                complete = complete + self.options[k] + "*atype*"

        newcomplete = complete.replace('"', "*doubleq*")
        self.options["ctext"] = newcomplete
        self.options["no_of_buttons"] = no_of_buttons

        if "caption" not in self.options:
            self.options["caption"] = ""
        else:
            self.options["caption"] = "data-caption='%s'" % self.options["caption"]

        if "include" not in self.options:
            self.options["include"] = ""
        else:
            lst = self.options["include"].split(",")
            lst = [x.strip() for x in lst]
            self.options["include"] = 'data-include="' + " ".join(lst) + '"'

        if "hidecode" in self.options:
            self.options["hidecode"] = 'data-hidecode="true"'
        else:
            self.options["hidecode"] = ""

        if "enabledownload" in self.options:
            self.options["enabledownload"] = 'data-enabledownload="true"'
        else:
            self.options["enabledownload"] = ""

        if "chatcodes" in self.options:
            self.options["chatcodes"] = 'data-chatcodes="true"'
        else:
            self.options["chatcodes"] = ""

        if "language" not in self.options:
            self.options["language"] = "python"

        if self.options["language"] == "html":
            self.options["language"] = "htmlmixed"
            self.options["initialcode"] = escape(self.options["initialcode"])

        if "nocodelens" in self.options or self.options["language"] not in [
            "python",
            "java",
            "c",
            "cpp",
        ]:
            self.options["codelens"] = ""
        else:
            self.options["codelens"] = 'data-codelens="true"'

        if "nopair" in self.options:
            self.options["nopair"] = 'data-nopair="true"'
        else:
            self.options["nopair"] = ""

        if "timelimit" not in self.options:
            self.options["timelimit"] = "data-timelimit=25000"
        else:
            self.options["timelimit"] = "data-timelimit=%s" % self.options["timelimit"]

        if "autorun" not in self.options:
            self.options["autorun"] = ""
        else:
            self.options["autorun"] = 'data-autorun="true"'

        if "coach" in self.options:
            self.options["coach"] = 'data-coach="true"'
        else:
            self.options["coach"] = ""

        # livecode options
        if "stdin" in self.options:
            self.options["stdin"] = "data-stdin='%s'" % self.options["stdin"]
        else:
            self.options["stdin"] = ""

        if "datafile" not in self.options:
            self.options["datafile"] = ""
        else:
            self.options["datafile"] = "data-datafile='%s'" % self.options["datafile"]

        if "sourcefile" not in self.options:
            self.options["sourcefile"] = ""
        else:
            self.options["sourcefile"] = (
                "data-sourcefile='%s'" % self.options["sourcefile"]
            )

        if "tie" in self.options:
            self.options["tie"] = "data-tie='{}'".format(self.options["tie"])
        else:
            self.options["tie"] = ""

        for opt, tp in [
            ("compileargs", "cargs"),
            ("linkargs", "largs"),
            ("runargs", "rargs"),
            ("interpreterargs", "iargs"),
        ]:
            if opt in self.options:
                self.options[tp] = 'data-{}="{}"'.format(opt, escape(self.options[opt]))
            else:
                self.options[tp] = ""

        # SQL Options
        if "dburl" in self.options:
            self.options["dburl"] = "data-dburl='{}'".format(self.options["dburl"])
        else:
            self.options["dburl"] = ""

        if "showlastsql" in self.options:
            self.options["showlastsql"] = 'data-showlastsql="true"'
        else:
            self.options["showlastsql"] = ""

        # other options

        if "gradebutton" not in self.options:
            self.options["gradebutton"] = ""
        else:
            self.options["gradebutton"] = "data-gradebutton=true"

        self.options["divclass"] = env.config.activecode_div_class
        if env.config.activecode_hide_load_history:
            self.options["hidehistory"] = "data-hidehistory=true"
        else:
            self.options["hidehistory"] = ""

        if env.config.wasm_uri:
            self.options["wasmuri"] = f"data-wasm={env.config.wasm_uri}"
        else:
            self.options["wasmuri"] = ""

        if self.content:
            if "^^^^" in self.content:
                idx = self.content.index("^^^^")
                prefix = "\n".join(self.content[:idx])
            else:
                prefix = ""
            if "====" in self.content:
                idx = self.content.index("====")
                source = "\n".join(self.content[:idx])
                suffix = "\n".join(self.content[idx + 1 :])
            else:
                source = "\n".join(self.content)
                suffix = "\n"
        else:
            source = "\n"
            suffix = "\n"
            prefix = "\n"

        self.options["xml_prefix_code"] = prefix
        self.options["xml_source_code"] = source
        self.options["xml_suffix_code"] = suffix

        course_name = env.config.html_context["course_id"]
        divid = self.options["divid"]

        engine, meta, sess = get_engine_meta()

        if engine:
            Source_code = Table(
                "source_code", meta, autoload=True, autoload_with=engine
            )
            engine.execute(
                Source_code.delete()
                .where(Source_code.c.acid == divid)
                .where(Source_code.c.course_id == course_name)
            )
            engine.execute(
                Source_code.insert().values(
                    acid=divid,
                    course_id=course_name,
                    main_code=source,
                    suffix_code=suffix,
                    includes=self.options["include"],
                    available_files=self.options.get("available_files", ""),
                )
            )
        else:
            if (
                not hasattr(env, "dberr_activecode_reported")
                or not env.dberr_activecode_reported
            ):
                env.dberr_activecode_reported = True
                print(
                    "Unable to save to source_code table in activecode.py. Possible problems:"
                )
                print("  1. dburl or course_id are not set in conf.py for your book")
                print("  2. unable to connect to the database using dburl")
                print("")
                print(
                    "This should only affect the grading interface. Everything else should be fine."
                )

        acnode = ActivecodeNode()
        acnode["runestone_options"] = self.options
        acnode["source"], acnode["line"] = self.state_machine.get_source_and_line(
            self.lineno
        )
        self.add_name(acnode)  # make this divid available as a target for :ref:

        maybeAddToAssignment(self)
        if explain_text:
            self.state.nested_parse(explain_text, self.content_offset, acnode)

        return [acnode]


class ActiveExercise(ActiveCode):
    required_arguments = 1
    optional_arguments = 0
    has_content = True

    def run(self):
        self.options["hidecode"] = "data-hidecode=true"
        self.options["gradebutton"] = "data-gradebutton=true"
        self.options["coach"] = "data-coach=true"
        return super(ActiveExercise, self).run()


if __name__ == "__main__":
    a = ActiveCode()



# Online Python Tutor
# https://github.com/pgbovine/OnlinePythonTutor/
#
# Copyright (C) Philip J. Guo (philip@pgbovine.net)
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


# This is the meat of the Online Python Tutor back-end.  It implements a
# full logger for Python program execution (based on pdb, the standard
# Python debugger imported via the bdb module), printing out the values
# of all in-scope data structures after each executed instruction.


import json
import random
from .pg_encoder import is_class, is_instance, ObjectEncoder, create_lambda_line_number
import sys
import bdb  # the KEY import here!
import re
import traceback
import types

# TODO: use the 'six' package to smooth out Py2 and Py3 differences
is_python3 = sys.version_info[0] == 3

# NB: don't use cStringIO since it doesn't support unicode!!!
if is_python3:
    import io as StringIO
    import io  # expose regular io for Python3 users too
else:
    import StringIO


# TODO: not threadsafe:

# upper-bound on the number of executed lines, in order to guard against
# infinite loops
# MAX_EXECUTED_LINES = 300
MAX_EXECUTED_LINES = 1000  # on 2016-05-01, I increased the limit from 300 to 1000 for Python due to popular user demand! and I also improved the warning message

# DEBUG = False
DEBUG = True

BREAKPOINT_STR = "#break"

CLASS_RE = re.compile(r"class\s+")


# simple sandboxing scheme:
#
# - use resource.setrlimit to deprive this process of ANY file descriptors
#   (which will cause file read/write and subprocess shell launches to fail)
# - restrict user builtins and module imports
#   (beware that this is NOT foolproof at all ... there are known flaws!)
#
# ALWAYS use defense-in-depth and don't just rely on these simple mechanisms
try:
    import resource

    resource_module_loaded = True
except ImportError:
    # Google App Engine doesn't seem to have the 'resource' module
    resource_module_loaded = False


# From http://coreygoldberg.blogspot.com/2009/05/python-redirect-or-turn-off-stdout-and.html
class NullDevice:
    def write(self, s):
        pass


# These could lead to XSS or other code injection attacks, so be careful:
__html__ = None


def setHTML(htmlStr):
    global __html__
    __html__ = htmlStr


__css__ = None


def setCSS(cssStr):
    global __css__
    __css__ = cssStr


__js__ = None


def setJS(jsStr):
    global __js__
    __js__ = jsStr


# ugh, I can't figure out why in Python 2, __builtins__ seems to
# be a dict, but in Python 3, __builtins__ seems to be a module,
# so just handle both cases ... UGLY!
if type(__builtins__) is dict:
    BUILTIN_IMPORT = __builtins__["__import__"]
else:
    assert type(__builtins__) is types.ModuleType
    BUILTIN_IMPORT = __builtins__.__import__


# whitelist of module imports
ALLOWED_STDLIB_MODULE_IMPORTS = (
    "math",
    "random",
    "time",
    "datetime",
    "functools",
    "itertools",
    "operator",
    "string",
    "collections",
    "re",
    "json",
    "heapq",
    "bisect",
    "copy",
    "hashlib",
)

# allow users to import but don't explicitly import it since it's
# already been done above
OTHER_STDLIB_WHITELIST = ("StringIO", "io")

# whitelist of custom modules to import into OPT
# (TODO: support modules in a subdirectory, but there are various
# logistical problems with doing so that I can't overcome at the moment,
# especially getting setHTML, setCSS, and setJS to work in the imported
# modules.)
CUSTOM_MODULE_IMPORTS = (
    # "runestone.codelens.callback_module",
    # "runestone.codelens.ttt_module",
    # "runestone.codelens.html_module",
    # "runestone.codelens.htmlexample_module",
    # # ignore these troublesome imports for now
    # #                         'watch_module',   # 'import sys' might be troublesome
    # #                         'bintree_module',
    # #                         'GChartWrapper',
    # "runestone.codelens.matrix",
    # "runestone.codelens.htmlFrame",
)


# PREEMPTIVELY import all of these modules, so that when the user's
# script imports them, it won't try to do a file read (since they've
# already been imported and cached in memory). Remember that when
# the user's code runs, resource.setrlimit(resource.RLIMIT_NOFILE, (0, 0))
# will already be in effect, so no more files can be opened.
#
# NB: All modules in CUSTOM_MODULE_IMPORTS will be imported, warts and
# all, so they better work on Python 2 and 3!
for m in ALLOWED_STDLIB_MODULE_IMPORTS + CUSTOM_MODULE_IMPORTS:
    __import__(m)


# Restrict imports to a whitelist
def __restricted_import__(*args):
    # filter args to ONLY take in real strings so that someone can't
    # subclass str and bypass the 'in' test on the next line
    args = [e for e in args if type(e) is str]

    if (
        args[0]
        in ALLOWED_STDLIB_MODULE_IMPORTS
        + CUSTOM_MODULE_IMPORTS
        + OTHER_STDLIB_WHITELIST
    ):
        imported_mod = BUILTIN_IMPORT(*args)

        if args[0] in CUSTOM_MODULE_IMPORTS:
            # add special magical functions to custom imported modules
            setattr(imported_mod, "setHTML", setHTML)
            setattr(imported_mod, "setCSS", setCSS)
            setattr(imported_mod, "setJS", setJS)

        # somewhat weak protection against imported modules that contain one
        # of these troublesome builtins. again, NOTHING is foolproof ...
        # just more defense in depth :)
        for mod in ("os", "sys", "posix", "gc"):
            if hasattr(imported_mod, mod):
                delattr(imported_mod, mod)

        return imported_mod
    else:
        raise ImportError("{0} not supported".format(args[0]))


# Support interactive user input by:
#
# 1. running the entire program up to a call to raw_input (or input in py3),
# 2. bailing and returning a trace ending in a special 'raw_input' event,
# 3. letting the web frontend issue a prompt to the user to grab a string,
# 4. RE-RUNNING the whole program with that string added to input_string_queue,
# 5. which should bring execution to the next raw_input call (if
#    available), or to termination.
# Repeat until no more raw_input calls are encountered.
# Note that this is mad inefficient, but is simple to implement!

# VERY IMPORTANT -- set random seed to 0 to ensure deterministic execution:

random.seed(0)

# queue of input strings passed from either raw_input or mouse_input
input_string_queue = []


def open_wrapper(*args):
    if is_python3:
        raise Exception(
            """open() is not supported by Python Tutor.
Instead use io.StringIO() to simulate a file.
Here is an example: http://goo.gl/uNvBGl"""
        )
    else:
        raise Exception(
            """open() is not supported by Python Tutor.
Instead use StringIO.StringIO() to simulate a file.
Here is an example: http://goo.gl/Q9xQ4p"""
        )


# create a more sensible error message for unsupported features
def create_banned_builtins_wrapper(fn_name):
    def err_func(*args):
        raise Exception("'" + fn_name + "' is not supported by Python Tutor.")

    return err_func


class RawInputException(Exception):
    pass


def raw_input_wrapper(prompt=""):
    if input_string_queue:
        input_str = input_string_queue.pop(0)

        # write the prompt and user input to stdout, to emulate what happens
        # at the terminal
        sys.stdout.write(str(prompt))  # always convert prompt into a string
        sys.stdout.write(input_str + "\n")  # newline to simulate the user hitting Enter
        return input_str
    raise RawInputException(str(prompt))  # always convert prompt into a string


# Python 2 input() does eval(raw_input())
def python2_input_wrapper(prompt=""):
    if input_string_queue:
        input_str = input_string_queue.pop(0)

        # write the prompt and user input to stdout, to emulate what happens
        # at the terminal
        sys.stdout.write(str(prompt))  # always convert prompt into a string
        sys.stdout.write(input_str + "\n")  # newline to simulate the user hitting Enter
        return eval(input_str)  # remember to eval!
    raise RawInputException(str(prompt))  # always convert prompt into a string


class MouseInputException(Exception):
    pass


def mouse_input_wrapper(prompt=""):
    if input_string_queue:
        return input_string_queue.pop(0)
    raise MouseInputException(prompt)


# blacklist of builtins
BANNED_BUILTINS = [
    "reload",
    "open",
    "compile",
    "file",
    "eval",
    "exec",
    "execfile",
    "exit",
    "quit",
    "help",
    "dir",
    "globals",
    "locals",
    "vars",
]
# Peter says 'apply' isn't dangerous, so don't ban it

IGNORE_VARS = set(
    (
        "__user_stdout__",
        "__OPT_toplevel__",
        "__builtins__",
        "__name__",
        "__exception__",
        "__doc__",
        "__package__",
    )
)


def get_user_stdout(frame):
    my_user_stdout = frame.f_globals["__user_stdout__"]

    # This is SUPER KRAZY! In Python 2, the buflist inside of a StringIO
    # instance can be made up of both str and unicode, so we need to convert
    # the str to unicode and replace invalid characters with the Unicode '?'
    # But leave unicode elements alone. This way, EVERYTHING inside buflist
    # will be unicode. (Note that in Python 3, everything is already unicode,
    # so we're fine.)
    if not is_python3:
        my_user_stdout.buflist = [
            (e.decode("utf-8", "replace") if type(e) is str else e)
            for e in my_user_stdout.buflist
        ]
    return my_user_stdout.getvalue()


"""
2013-12-26

Okay, what's with this f_valuestack business?

If you compile your own CPython and patch Objects/frameobject.c to add a
Python accessor for f_valuestack, then you can actually access the value
stack, which is useful for, say, grabbbing the objects within
list/set/dict comprehensions as they're being built. e.g., try:

    z = [x*y for x in range(5) for y in range(5)]

Note that on pythontutor.com, I am currently running custom-compiled
versions of Python-2.7.6 and Python-3.3.3 with this f_valuestack hack.
Unless you run your own custom CPython, you won't get these benefits.


Patch:

 static PyObject *
 frame_getlineno(PyFrameObject *f, void *closure)
 {
     return PyLong_FromLong(PyFrame_GetLineNumber(f));
 }

+// copied from Py2crazy, which was for Python 2, but let's hope this still works!
+static PyObject *
+frame_getvaluestack(PyFrameObject* f) {
+    // pgbovine - TODO: will this memory leak? hopefully not,
+    // since all other accessors seem to follow the same idiom
+    PyObject* lst = PyList_New(0);
+    if (f->f_stacktop != NULL) {
+        PyObject** p = NULL;
+        for (p = f->f_valuestack; p < f->f_stacktop; p++) {
+            PyList_Append(lst, *p);
+        }
+    }
+
+    return lst;
+}
+
 /* Setter for f_lineno - you can set f_lineno from within a trace function in
  * order to jump to a given line of code, subject to some restrictions.  Most
  * lines are OK to jump to because they don't make any assumptions about the
@@ -368,6 +384,11 @@

 static PyGetSetDef frame_getsetlist[] = {
     {"f_locals",        (getter)frame_getlocals, NULL, NULL},
     {"f_lineno",        (getter)frame_getlineno,
                     (setter)frame_setlineno, NULL},
     {"f_trace",         (getter)frame_gettrace, (setter)frame_settrace, NULL},
+
+    // pgbovine
+    {"f_valuestack",(getter)frame_getvaluestack,
+                    (setter)NULL /* don't let it be set */, NULL},
+
     {0}
 };
"""

# at_global_scope should be true only if 'frame' represents the global scope


def get_user_globals(frame, at_global_scope=False):
    d = filter_var_dict(frame.f_globals)

    # don't blurt out all of f_valuestack for now ...
    """
  if at_global_scope and hasattr(frame, 'f_valuestack'):
    for (i, e) in enumerate(frame.f_valuestack):
      d['_tmp' + str(i+1)] = e
  """

    # print out list objects being built up in Python 2.x list comprehensions
    # (which don't have its own special <listcomp> frame, sadly)
    if not is_python3 and hasattr(frame, "f_valuestack"):
        for (i, e) in enumerate([e for e in frame.f_valuestack if type(e) is list]):
            d["_tmp" + str(i + 1)] = e

    # also filter out __return__ for globals only, but NOT for locals
    if "__return__" in d:
        del d["__return__"]
    return d


def get_user_locals(frame):
    ret = filter_var_dict(frame.f_locals)
    # don't blurt out all of f_valuestack for now ...
    """
  if hasattr(frame, 'f_valuestack'):
    for (i, e) in enumerate(frame.f_valuestack):
      ret['_tmp' + str(i+1)] = e
  """

    # special printing of list/set/dict comprehension objects as they are
    # being built up incrementally ...
    f_name = frame.f_code.co_name
    if hasattr(frame, "f_valuestack"):
        # print out list objects being built up in Python 2.x list comprehensions
        # (which don't have its own special <listcomp> frame, sadly)
        if not is_python3:
            for (i, e) in enumerate([e for e in frame.f_valuestack if type(e) is list]):
                ret["_tmp" + str(i + 1)] = e

        # for dict and set comprehensions, which have their own frames:
        if f_name.endswith("comp>"):
            for (i, e) in enumerate(
                [e for e in frame.f_valuestack if type(e) in (list, set, dict)]
            ):
                ret["_tmp" + str(i + 1)] = e

    return ret


def filter_var_dict(d):
    ret = {}
    for (k, v) in d.items():
        if k not in IGNORE_VARS:
            ret[k] = v
    return ret


# yield all function objects locally-reachable from frame,
# making sure to traverse inside all compound objects ...
def visit_all_locally_reachable_function_objs(frame):
    for (k, v) in get_user_locals(frame).items():
        for e in visit_function_obj(v, set()):
            if e:  # only non-null if it's a function object
                assert type(e) in (types.FunctionType, types.MethodType)
                yield e


# TODO: this might be slow if we're traversing inside lots of objects:
def visit_function_obj(v, ids_seen_set):
    v_id = id(v)

    # to prevent infinite loop
    if v_id in ids_seen_set:
        yield None
    else:
        ids_seen_set.add(v_id)

        typ = type(v)

        # simple base case
        if typ in (types.FunctionType, types.MethodType):
            yield v

        # recursive cases
        elif typ in (list, tuple, set):
            for child in v:
                for child_res in visit_function_obj(child, ids_seen_set):
                    yield child_res

        elif typ == dict or is_class(v) or is_instance(v):
            contents_dict = None

            if typ == dict:
                contents_dict = v
            # warning: some classes or instances don't have __dict__ attributes
            elif hasattr(v, "__dict__"):
                contents_dict = v.__dict__

            if contents_dict:
                for (key_child, val_child) in contents_dict.items():
                    for key_child_res in visit_function_obj(key_child, ids_seen_set):
                        yield key_child_res
                    for val_child_res in visit_function_obj(val_child, ids_seen_set):
                        yield val_child_res

        # degenerate base case
        yield None


class PGLogger(bdb.Bdb):
    def __init__(
        self,
        cumulative_mode,
        heap_primitives,
        show_only_outputs,
        finalizer_func,
        disable_security_checks=False,
        crazy_mode=False,
    ):
        bdb.Bdb.__init__(self)
        self.mainpyfile = ""
        self._wait_for_mainpyfile = 0

        self.disable_security_checks = disable_security_checks

        # if True, then displays ALL stack frames that have ever existed
        # rather than only those currently on the stack (and their
        # lexical parents)
        self.cumulative_mode = cumulative_mode

        # if True, then render certain primitive objects as heap objects
        self.render_heap_primitives = heap_primitives

        # if True, then don't render any data structures in the trace,
        # and show only outputs
        self.show_only_outputs = show_only_outputs

        # Run using the custom Py2crazy Python interpreter
        self.crazy_mode = crazy_mode

        # a function that takes the output trace as a parameter and
        # processes it
        self.finalizer_func = finalizer_func

        # each entry contains a dict with the information for a single
        # executed line
        self.trace = []

        # if this is true, don't put any more stuff into self.trace
        self.done = False

        # if this is non-null, don't do any more tracing until a
        # 'return' instruction with a stack gotten from
        # get_stack_code_IDs() that matches wait_for_return_stack
        self.wait_for_return_stack = None

        # http://stackoverflow.com/questions/2112396/in-python-in-google-app-engine-how-do-you-capture-output-produced-by-the-print
        self.GAE_STDOUT = sys.stdout

        # Key:   function object
        # Value: parent frame
        self.closures = {}

        # Key:   code object for a lambda
        # Value: parent frame
        self.lambda_closures = {}

        # set of function objects that were defined in the global scope
        self.globally_defined_funcs = set()

        # Key: frame object
        # Value: monotonically increasing small ID, based on call order
        self.frame_ordered_ids = {}
        self.cur_frame_id = 1

        # List of frames to KEEP AROUND after the function exits.
        # If cumulative_mode is True, then keep ALL frames in
        # zombie_frames; otherwise keep only frames where
        # nested functions were defined within them.
        self.zombie_frames = []

        # set of elements within zombie_frames that are also
        # LEXICAL PARENTS of other frames
        self.parent_frames_set = set()

        # all globals that ever appeared in the program, in the order in
        # which they appeared. note that this might be a superset of all
        # the globals that exist at any particular execution point,
        # since globals might have been deleted (using, say, 'del')
        self.all_globals_in_order = []

        # very important for this single object to persist throughout
        # execution, or else canonical small IDs won't be consistent.
        self.encoder = ObjectEncoder(self.render_heap_primitives)

        self.executed_script = None  # Python script to be executed!

        # if there is at least one line that ends with BREAKPOINT_STR,
        # then activate "breakpoint mode", where execution should stop
        # ONLY at breakpoint lines.
        self.breakpoints = []

        self.prev_lineno = -1  # keep track of previous line just executed

    def get_frame_id(self, cur_frame):
        return self.frame_ordered_ids[cur_frame]

    # Returns the (lexical) parent of a function value.
    def get_parent_of_function(self, val):
        if val in self.closures:
            return self.get_frame_id(self.closures[val])
        elif val in self.lambda_closures:
            return self.get_frame_id(self.lambda_closures[val])
        else:
            return None

    # Returns the (lexical) parent frame of the function that was called
    # to create the stack frame 'frame'.
    #
    # OKAY, this is a SUPER hack, but I don't see a way around it
    # since it's impossible to tell exactly which function
    # ('closure') object was called to create 'frame'.
    #
    # The Python interpreter doesn't maintain this information,
    # so unless we hack the interpreter, we will simply have
    # to make an educated guess based on the contents of local
    # variables inherited from possible parent frame candidates.
    def get_parent_frame(self, frame):
        # print >> sys.stderr, 'get_parent_frame: frame.f_code', frame.f_code
        for (func_obj, parent_frame) in self.closures.items():
            # ok, there's a possible match, but let's compare the
            # local variables in parent_frame to those of frame
            # to make sure. this is a hack that happens to work because in
            # Python, each stack frame inherits ('inlines') a copy of the
            # variables from its (lexical) parent frame.
            if func_obj.__code__ == frame.f_code:
                all_matched = True
                for k in frame.f_locals:
                    # Do not try to match local names
                    if k in frame.f_code.co_varnames:
                        continue
                    if k != "__return__" and k in parent_frame.f_locals:
                        if parent_frame.f_locals[k] != frame.f_locals[k]:
                            all_matched = False
                            break

                if all_matched:
                    return parent_frame

        for (lambda_code_obj, parent_frame) in self.lambda_closures.items():
            if lambda_code_obj == frame.f_code:
                # TODO: should we do more verification like above?!?
                return parent_frame

        return None

    def lookup_zombie_frame_by_id(self, frame_id):
        # TODO: kinda inefficient
        for e in self.zombie_frames:
            if self.get_frame_id(e) == frame_id:
                return e
        assert False  # should never get here

    # unused ...
    # def reset(self):
    #    bdb.Bdb.reset(self)
    #    self.forget()

    def forget(self):
        self.lineno = None
        self.stack = []
        self.curindex = 0
        self.curframe = None

    def setup(self, f, t):
        self.forget()
        self.stack, self.curindex = self.get_stack(f, t)
        self.curframe = self.stack[self.curindex][0]

    # should be a reasonably unique ID to match calls and returns:
    def get_stack_code_IDs(self):
        return [id(e[0].f_code) for e in self.stack]

    # Override Bdb methods

    def user_call(self, frame, argument_list):
        """This method is called when there is the remote possibility
        that we ever need to stop in this function."""
        # TODO: figure out a way to move this down to 'def interaction'
        # or right before self.trace.append ...
        if self.done:
            return

        if self._wait_for_mainpyfile:
            return
        if self.stop_here(frame):
            # delete __return__ so that on subsequent calls to
            # a generator function, the OLD yielded (returned)
            # value gets deleted from the frame ...
            try:
                del frame.f_locals["__return__"]
            except KeyError:
                pass

            self.interaction(frame, None, "call")

    def user_line(self, frame):
        """This function is called when we stop or break at this line."""
        if self.done:
            return

        if self._wait_for_mainpyfile:
            if (
                self.canonic(frame.f_code.co_filename) != "<string>"
                or frame.f_lineno <= 0
            ):
                return
            self._wait_for_mainpyfile = 0
        self.interaction(frame, None, "step_line")

    def user_return(self, frame, return_value):
        """This function is called when a return trap is set here."""
        if self.done:
            return

        frame.f_locals["__return__"] = return_value
        self.interaction(frame, None, "return")

    def user_exception(self, frame, exc_info):
        """This function is called if an exception occurs,
        but only if we are to stop at or just below this level."""
        if self.done:
            return

        exc_type, exc_value, exc_traceback = exc_info
        frame.f_locals["__exception__"] = exc_type, exc_value
        if type(exc_type) == type(""):
            exc_type_name = exc_type
        else:
            exc_type_name = exc_type.__name__

        if exc_type_name == "RawInputException":
            raw_input_arg = str(
                exc_value.args[0]
            )  # make sure it's a string so it's JSON serializable!
            self.trace.append(dict(event="raw_input", prompt=raw_input_arg))
            self.done = True
        elif exc_type_name == "MouseInputException":
            mouse_input_arg = str(
                exc_value.args[0]
            )  # make sure it's a string so it's JSON serializable!
            self.trace.append(dict(event="mouse_input", prompt=mouse_input_arg))
            self.done = True
        else:
            self.interaction(frame, exc_traceback, "exception")

    def get_script_line(self, n):
        return self.executed_script_lines[n - 1]

    # General interaction function

    def interaction(self, frame, traceback, event_type):
        self.setup(frame, traceback)
        tos = self.stack[self.curindex]
        top_frame = tos[0]
        lineno = tos[1]

        # debug ...
        """
        print >> sys.stderr
        print >> sys.stderr, '=== STACK ===', 'curindex:', self.curindex
        for (e,ln) in self.stack:
          print >> sys.stderr, e.f_code.co_name + ' ' + e.f_code.co_filename + ' ' + str(ln)
        print >> sys.stderr, "top_frame", top_frame.f_code.co_name, top_frame.f_code
        """

        # don't trace inside of ANY functions that aren't user-written code
        # (e.g., those from imported modules -- e.g., random, re -- or the
        # __restricted_import__ function in this file)
        #
        # empirically, it seems like the FIRST entry in self.stack is
        # the 'run' function from bdb.py, but everything else on the
        # stack is the user program's "real stack"

        # Look only at the "topmost" frame on the stack ...

        # it seems like user-written code has a filename of '<string>',
        # but maybe there are false positives too?
        if self.canonic(top_frame.f_code.co_filename) != "<string>":
            return
        # also don't trace inside of the magic "constructor" code
        if top_frame.f_code.co_name == "__new__":
            return
        # or __repr__, which is often called when running print statements
        if top_frame.f_code.co_name == "__repr__":
            return

        # if top_frame.f_globals doesn't contain the sentinel '__OPT_toplevel__',
        # then we're in another global scope altogether, so skip it!
        # (this comes up in tests/backend-tests/namedtuple.txt)
        if "__OPT_toplevel__" not in top_frame.f_globals:
            return

        # OLD CODE -- bail if any element on the stack matches these conditions
        # note that the old code passes tests/backend-tests/namedtuple.txt
        # but the new code above doesn't :/
        """
        for (cur_frame, cur_line) in self.stack[1:]:
          # it seems like user-written code has a filename of '<string>',
          # but maybe there are false positives too?
          if self.canonic(cur_frame.f_code.co_filename) != '<string>':
            return
          # also don't trace inside of the magic "constructor" code
          if cur_frame.f_code.co_name == '__new__':
            return
          # or __repr__, which is often called when running print statements
          if cur_frame.f_code.co_name == '__repr__':
            return
        """

        # don't trace if wait_for_return_stack is non-null ...
        if self.wait_for_return_stack:
            if event_type == "return" and (
                self.wait_for_return_stack == self.get_stack_code_IDs()
            ):
                self.wait_for_return_stack = None  # reset!
            return  # always bail!
        else:
            # Skip all "calls" that are actually class definitions, since
            # those faux calls produce lots of ugly cruft in the trace.
            #
            # NB: Only trigger on calls to functions defined in
            # user-written code (i.e., co_filename == '<string>'), but that
            # should already be ensured by the above check for whether we're
            # in user-written code.
            if event_type == "call":
                func_line = self.get_script_line(top_frame.f_code.co_firstlineno)
                if CLASS_RE.match(func_line.lstrip()):  # ignore leading spaces
                    self.wait_for_return_stack = self.get_stack_code_IDs()
                    return

        self.encoder.reset_heap()  # VERY VERY VERY IMPORTANT,
        # or else we won't properly capture heap object mutations in the trace!

        if event_type == "call":
            # Don't be so strict about this assertion because it FAILS
            # when you're calling a generator (not for the first time),
            # since that frame has already previously been on the stack ...
            # assert top_frame not in self.frame_ordered_ids

            self.frame_ordered_ids[top_frame] = self.cur_frame_id
            self.cur_frame_id += 1

            if self.cumulative_mode:
                self.zombie_frames.append(top_frame)

        # only render zombie frames that are NO LONGER on the stack
        #
        # subtle: self.stack[:self.curindex+1] is the real stack, since
        # everything after self.curindex+1 is beyond the top of the
        # stack. this seems to be relevant only when there's an exception,
        # since the ENTIRE stack is preserved but self.curindex
        # starts decrementing as the exception bubbles up the stack.
        cur_stack_frames = [e[0] for e in self.stack[: self.curindex + 1]]
        zombie_frames_to_render = [
            e for e in self.zombie_frames if e not in cur_stack_frames
        ]

        # each element is a pair of (function name, ENCODED locals dict)
        encoded_stack_locals = []

        # returns a dict with keys: function name, frame id, id of parent frame, encoded_locals dict
        def create_encoded_stack_entry(cur_frame):
            # print >> sys.stderr, '- create_encoded_stack_entry', cur_frame, self.closures, self.lambda_closures
            ret = {}

            parent_frame_id_list = []

            f = cur_frame
            while True:
                p = self.get_parent_frame(f)
                if p:
                    pid = self.get_frame_id(p)
                    assert pid
                    parent_frame_id_list.append(pid)
                    f = p
                else:
                    break

            cur_name = cur_frame.f_code.co_name

            if cur_name == "":
                cur_name = "unnamed function"

            # augment lambdas with line number
            if cur_name == "<lambda>":
                cur_name += create_lambda_line_number(
                    cur_frame.f_code, self.encoder.line_to_lambda_code
                )

            # encode in a JSON-friendly format now, in order to prevent ill
            # effects of aliasing later down the line ...
            encoded_locals = {}

            for (k, v) in get_user_locals(cur_frame).items():
                is_in_parent_frame = False

                # don't display locals that appear in your parents' stack frames,
                # since that's redundant
                for pid in parent_frame_id_list:
                    parent_frame = self.lookup_zombie_frame_by_id(pid)
                    if k in parent_frame.f_locals:
                        # ignore __return__, which is never copied
                        if k != "__return__":
                            # these values SHOULD BE ALIASES
                            # (don't do an 'is' check since it might not fire for primitives)
                            if parent_frame.f_locals[k] == v:
                                is_in_parent_frame = True

                if is_in_parent_frame and k not in cur_frame.f_code.co_varnames:
                    continue

                # don't display some built-in locals ...
                if k == "__module__":
                    continue

                encoded_val = self.encoder.encode(v, self.get_parent_of_function)
                encoded_locals[k] = encoded_val

            # order the variable names in a sensible way:

            # Let's start with co_varnames, since it (often) contains all
            # variables in this frame, some of which might not exist yet.
            ordered_varnames = []
            for e in cur_frame.f_code.co_varnames:
                if e in encoded_locals:
                    ordered_varnames.append(e)

            # sometimes co_varnames doesn't contain all of the true local
            # variables: e.g., when executing a 'class' definition.  in that
            # case, iterate over encoded_locals and push them onto the end
            # of ordered_varnames in alphabetical order
            for e in sorted(encoded_locals.keys()):
                if e != "__return__" and e not in ordered_varnames:
                    ordered_varnames.append(e)

            # finally, put __return__ at the very end
            if "__return__" in encoded_locals:
                ordered_varnames.append("__return__")

            # doctor Python 3 initializer to look like a normal function (denero)
            if "__locals__" in encoded_locals:
                ordered_varnames.remove("__locals__")
                local = encoded_locals.pop("__locals__")
                if encoded_locals.get("__return__", True) is None:
                    encoded_locals["__return__"] = local

            # crucial sanity checks!
            assert len(ordered_varnames) == len(encoded_locals)
            for e in ordered_varnames:
                assert e in encoded_locals

            return dict(
                func_name=cur_name,
                is_parent=(cur_frame in self.parent_frames_set),
                frame_id=self.get_frame_id(cur_frame),
                parent_frame_id_list=parent_frame_id_list,
                encoded_locals=encoded_locals,
                ordered_varnames=ordered_varnames,
            )

        i = self.curindex

        # look for whether a nested function has been defined during
        # this particular call:
        if i > 1:  # i == 1 implies that there's only a global scope visible
            for v in visit_all_locally_reachable_function_objs(top_frame):
                if v not in self.closures and v not in self.globally_defined_funcs:

                    # Look for the presence of the code object (v.func_code
                    # for Python 2 or v.__code__ for Python 3) in the
                    # constant pool (f_code.co_consts) of an enclosing
                    # stack frame, and set that frame as your parent.
                    #
                    # This technique properly handles lambdas passed as
                    # function parameters. e.g., this example:
                    #
                    # def foo(x):
                    #   bar(lambda y: x + y)
                    # def bar(a):
                    #   print a(20)
                    # foo(10)
                    chosen_parent_frame = None
                    # SUPER hacky but seems to work -- use reversed(self.stack)
                    # because we want to traverse starting from the TOP of the stack
                    # (most recent frame) and find the first frame containing
                    # a constant code object that matches v.__code__ or v.func_code
                    #
                    # required for this example from Berkeley CS61a:
                    #
                    # def f(p, k):
                    #     def g():
                    #         print(k)
                    #     if k == 0:
                    #         f(g, 1)
                    # f(None, 0)
                    #
                    # there are two calls to f, each of which defines a
                    # closure g that should point to the respective frame.
                    #
                    # note that for the second call to f, the parent of the
                    # g defined in there should be that frame, which is at
                    # the TOP of the stack. this reversed() hack does the
                    # right thing. note that if you don't traverse the stack
                    # backwards, then you will mistakenly get the parent as
                    # the FIRST f frame (bottom of the stack).
                    for (my_frame, my_lineno) in reversed(self.stack):
                        if chosen_parent_frame:
                            break

                        for frame_const in my_frame.f_code.co_consts:
                            if frame_const is (
                                v.__code__ if is_python3 else v.func_code
                            ):
                                chosen_parent_frame = my_frame
                                break

                    # 2013-12-01 commented out this line so tests/backend-tests/papajohn-monster.txt
                    # works without an assertion failure ...
                    # assert chosen_parent_frame # I hope this always passes :0

                    # this condition should be False for functions declared in global scope ...
                    if chosen_parent_frame in self.frame_ordered_ids:
                        self.closures[v] = chosen_parent_frame
                        self.parent_frames_set.add(
                            chosen_parent_frame
                        )  # unequivocally add to this set!!!
                        if not chosen_parent_frame in self.zombie_frames:
                            self.zombie_frames.append(chosen_parent_frame)
            else:
                # look for code objects of lambdas defined within this
                # function, which comes up in cases like line 2 of:
                # def x(y):
                #   (lambda z: lambda w: z+y)(y)
                #
                # x(42)
                if top_frame.f_code.co_consts:
                    for e in top_frame.f_code.co_consts:
                        if type(e) == types.CodeType and e.co_name == "<lambda>":
                            # TODO: what if it's already in lambda_closures?
                            self.lambda_closures[e] = top_frame
                            self.parent_frames_set.add(
                                top_frame
                            )  # copy-paste from above
                            if not top_frame in self.zombie_frames:
                                self.zombie_frames.append(top_frame)
        else:
            # if there is only a global scope visible ...
            for (k, v) in get_user_globals(top_frame).items():
                if (
                    type(v) in (types.FunctionType, types.MethodType)
                    and v not in self.closures
                ):
                    self.globally_defined_funcs.add(v)

        # climb up until you find '<module>', which is (hopefully) the global scope
        while True:
            cur_frame = self.stack[i][0]
            cur_name = cur_frame.f_code.co_name
            if cur_name == "<module>":
                break

            # do this check because in some cases, certain frames on the
            # stack might NOT be tracked, so don't push a stack entry for
            # those frames. this happens when you have a callback function
            # in an imported module. e.g., your code:
            #     def foo():
            #         bar(baz)
            #
            #     def baz(): pass
            #
            # imported module code:
            #     def bar(callback_func):
            #         callback_func()
            #
            # when baz is executing, the real stack is [foo, bar, baz] but
            # bar is in imported module code, so pg_logger doesn't trace
            # it, and it doesn't show up in frame_ordered_ids. thus, the
            # stack to render should only be [foo, baz].
            if cur_frame in self.frame_ordered_ids:
                encoded_stack_locals.append(create_encoded_stack_entry(cur_frame))
            i -= 1

        zombie_encoded_stack_locals = [
            create_encoded_stack_entry(e) for e in zombie_frames_to_render
        ]

        # encode in a JSON-friendly format now, in order to prevent ill
        # effects of aliasing later down the line ...
        encoded_globals = {}
        for (k, v) in get_user_globals(
            tos[0], at_global_scope=(self.curindex <= 1)
        ).items():
            encoded_val = self.encoder.encode(v, self.get_parent_of_function)
            encoded_globals[k] = encoded_val

            if k not in self.all_globals_in_order:
                self.all_globals_in_order.append(k)

        # filter out globals that don't exist at this execution point
        # (because they've been, say, deleted with 'del')
        ordered_globals = [e for e in self.all_globals_in_order if e in encoded_globals]
        assert len(ordered_globals) == len(encoded_globals)

        # merge zombie_encoded_stack_locals and encoded_stack_locals
        # into one master ordered list using some simple rules for
        # making it look aesthetically pretty
        stack_to_render = []

        # first push all regular stack entries
        if encoded_stack_locals:
            for e in encoded_stack_locals:
                e["is_zombie"] = False
                e["is_highlighted"] = False
                stack_to_render.append(e)

            # highlight the top-most active stack entry
            stack_to_render[0]["is_highlighted"] = True

        # now push all zombie stack entries
        for e in zombie_encoded_stack_locals:
            # don't display return value for zombie frames
            # TODO: reconsider ...
            """
          try:
            e['ordered_varnames'].remove('__return__')
          except ValueError:
            pass
          """

            e["is_zombie"] = True
            e["is_highlighted"] = False  # never highlight zombie entries

            stack_to_render.append(e)

        # now sort by frame_id since that sorts frames in "chronological
        # order" based on the order they were invoked
        stack_to_render.sort(key=lambda e: e["frame_id"])

        # create a unique hash for this stack entry, so that the
        # frontend can uniquely identify it when doing incremental
        # rendering. the strategy is to use a frankenstein-like mix of the
        # relevant fields to properly disambiguate closures and recursive
        # calls to the same function
        for e in stack_to_render:
            hash_str = e["func_name"]
            # frame_id is UNIQUE, so it can disambiguate recursive calls
            hash_str += "_f" + str(e["frame_id"])

            # needed to refresh GUI display ...
            if e["is_parent"]:
                hash_str += "_p"

            # TODO: this is no longer needed, right? (since frame_id is unique)
            # if e['parent_frame_id_list']:
            #  hash_str += '_p' + '_'.join([str(i) for i in e['parent_frame_id_list']])
            if e["is_zombie"]:
                hash_str += "_z"

            e["unique_hash"] = hash_str

        if self.show_only_outputs:
            trace_entry = dict(
                line=lineno,
                event=event_type,
                func_name=tos[0].f_code.co_name,
                globals={},
                ordered_globals=[],
                stack_to_render=[],
                heap={},
                stdout=get_user_stdout(tos[0]),
            )
        else:
            trace_entry = dict(
                line=lineno,
                event=event_type,
                func_name=tos[0].f_code.co_name,
                globals=encoded_globals,
                ordered_globals=ordered_globals,
                stack_to_render=stack_to_render,
                heap=self.encoder.get_heap(),
                stdout=get_user_stdout(tos[0]),
            )

        # optional column numbers for greater precision
        # (only relevant in Py2crazy, a hacked CPython that supports column numbers)
        if self.crazy_mode:
            # at the very least, grab the column number
            trace_entry["column"] = frame.f_colno

            # now try to find start_col and extent
            # (-1 is an invalid instruction index)
            if frame.f_lasti >= 0:
                key = (
                    frame.f_code.co_code,
                    frame.f_lineno,
                    frame.f_colno,
                    frame.f_lasti,
                )
                if key in self.bytecode_map:
                    v = self.bytecode_map[key]
                    trace_entry["expr_start_col"] = v.start_col
                    trace_entry["expr_width"] = v.extent
                    trace_entry["opcode"] = v.opcode

        # TODO: refactor into a non-global
        global __html__, __css__, __js__
        if __html__:
            trace_entry["html_output"] = __html__
        if __css__:
            trace_entry["css_output"] = __css__
        if __js__:
            trace_entry["js_output"] = __js__

        # if there's an exception, then record its info:
        if event_type == "exception":
            # always check in f_locals
            exc = frame.f_locals["__exception__"]
            trace_entry["exception_msg"] = exc[0].__name__ + ": " + str(exc[1])

        # append to the trace only the breakpoint line and the next
        # executed line, so that if you set only ONE breakpoint, OPT shows
        # the state before and after that line gets executed.
        append_to_trace = True
        if self.breakpoints:
            if not (
                (lineno in self.breakpoints) or (self.prev_lineno in self.breakpoints)
            ):
                append_to_trace = False

            # TRICKY -- however, if there's an exception, then ALWAYS
            # append it to the trace, so that the error can be displayed
            if event_type == "exception":
                append_to_trace = True

        self.prev_lineno = lineno

        if append_to_trace:
            self.trace.append(trace_entry)

        # sanity check to make sure the state of the world at a 'call' instruction
        # is identical to that at the instruction immediately following it ...
        """
        if len(self.trace) > 1:
          cur = self.trace[-1]
          prev = self.trace[-2]
          if prev['event'] == 'call':
            assert cur['globals'] == prev['globals']
            for (s1, s2) in zip(cur['stack_to_render'], prev['stack_to_render']):
              assert s1 == s2
            assert cur['heap'] == prev['heap']
            assert cur['stdout'] == prev['stdout']
        """

        if len(self.trace) >= MAX_EXECUTED_LINES:
            self.trace.append(
                dict(
                    event="instruction_limit_reached",
                    exception_msg="Stopped after running "
                    + str(MAX_EXECUTED_LINES)
                    + " steps. Please shorten your code,\nsince Python Tutor is not designed to handle long-running code.",
                )
            )
            self.force_terminate()

        self.forget()

    def _runscript(self, script_str, custom_globals=None):
        self.executed_script = script_str
        self.executed_script_lines = self.executed_script.splitlines()

        for (i, line) in enumerate(self.executed_script_lines):
            line_no = i + 1
            if line.endswith(BREAKPOINT_STR):
                self.breakpoints.append(line_no)

        # populate an extent map to get more accurate ranges from code
        if self.crazy_mode:
            # in Py2crazy standard library as Python-2.7.5/Lib/super_dis.py
            import super_dis

            try:
                self.bytecode_map = super_dis.get_bytecode_map(self.executed_script)
            except:
                # failure oblivious
                self.bytecode_map = {}

        # When bdb sets tracing, a number of call and line events happens
        # BEFORE debugger even reaches user's code (and the exact sequence of
        # events depends on python version). So we take special measures to
        # avoid stopping before we reach the main script (see user_line and
        # user_call for details).
        self._wait_for_mainpyfile = 1

        # ok, let's try to sorta 'sandbox' the user script by not
        # allowing certain potentially dangerous operations.
        user_builtins = {}

        # ugh, I can't figure out why in Python 2, __builtins__ seems to
        # be a dict, but in Python 3, __builtins__ seems to be a module,
        # so just handle both cases ... UGLY!
        if type(__builtins__) is dict:
            builtin_items = __builtins__.items()
        else:
            assert type(__builtins__) is types.ModuleType
            builtin_items = []
            for k in dir(__builtins__):
                builtin_items.append((k, getattr(__builtins__, k)))

        for (k, v) in builtin_items:
            if k == "open":  # put this before BANNED_BUILTINS
                user_builtins[k] = open_wrapper
            elif k in BANNED_BUILTINS:
                user_builtins[k] = create_banned_builtins_wrapper(k)
            elif k == "__import__":
                user_builtins[k] = __restricted_import__
            else:
                if k == "raw_input":
                    user_builtins[k] = raw_input_wrapper
                elif k == "input":
                    if is_python3:
                        # Python 3 input() is Python 2 raw_input()
                        user_builtins[k] = raw_input_wrapper
                    else:
                        user_builtins[k] = python2_input_wrapper
                else:
                    user_builtins[k] = v

        user_builtins["mouse_input"] = mouse_input_wrapper

        # TODO: we can disable these imports here, but a crafty user can
        # always get a hold of them by importing one of the external
        # modules, so there's no point in trying security by obscurity
        user_builtins["setHTML"] = setHTML
        user_builtins["setCSS"] = setCSS
        user_builtins["setJS"] = setJS

        user_stdout = StringIO.StringIO()

        sys.stdout = user_stdout

        self.ORIGINAL_STDERR = sys.stderr

        # don't do this, or else certain kinds of errors, such as syntax
        # errors, will be silently ignored. WEIRD!
        # sys.stderr = NullDevice # silence errors

        user_globals = {
            "__name__": "__main__",
            "__builtins__": user_builtins,
            "__user_stdout__": user_stdout,
            # sentinel value for frames deriving from a top-level module
            "__OPT_toplevel__": True,
        }

        if custom_globals:
            user_globals.update(custom_globals)

        try:
            # enforce resource limits RIGHT BEFORE running script_str

            # set ~200MB virtual memory limit AND a 5-second CPU time
            # limit (tuned for Webfaction shared hosting) to protect against
            # memory bombs such as:
            #   x = 2
            #   while True: x = x*x
            if resource_module_loaded and (not self.disable_security_checks):
                resource.setrlimit(resource.RLIMIT_AS, (200000000, 200000000))
                resource.setrlimit(resource.RLIMIT_CPU, (5, 5))

                # protect against unauthorized filesystem accesses ...
                resource.setrlimit(
                    resource.RLIMIT_NOFILE, (0, 0)
                )  # no opened files allowed

                # VERY WEIRD. If you activate this resource limitation, it
                # ends up generating an EMPTY trace for the following program:
                #   "x = 0\nfor i in range(10):\n  x += 1\n   print x\n  x += 1\n"
                # (at least on my Webfaction hosting with Python 2.7)
                # resource.setrlimit(resource.RLIMIT_FSIZE, (0, 0))  # (redundancy for paranoia)

                # The posix module is a built-in and has a ton of OS access
                # facilities ... if you delete those functions from
                # sys.modules['posix'], it seems like they're gone EVEN IF
                # someone else imports posix in a roundabout way. Of course,
                # I don't know how foolproof this scheme is, though.
                # (It's not sufficient to just "del sys.modules['posix']";
                #  it can just be reimported without accessing an external
                #  file and tripping RLIMIT_NOFILE, since the posix module
                #  is baked into the python executable, ergh. Actually DON'T
                #  "del sys.modules['posix']", since re-importing it will
                #  refresh all of the attributes. ergh^2)
                for a in dir(sys.modules["posix"]):
                    delattr(sys.modules["posix"], a)
                # do the same with os
                for a in dir(sys.modules["os"]):
                    # 'path' is needed for __restricted_import__ to work
                    # and 'stat' is needed for some errors to be reported properly
                    if a not in ("path", "stat"):
                        delattr(sys.modules["os"], a)
                # ppl can dig up trashed objects with gc.get_objects()
                import gc

                for a in dir(sys.modules["gc"]):
                    delattr(sys.modules["gc"], a)
                del sys.modules["gc"]

                # sys.modules contains an in-memory cache of already-loaded
                # modules, so if you delete modules from here, they will
                # need to be re-loaded from the filesystem.
                #
                # Thus, as an extra precaution, remove these modules so that
                # they can't be re-imported without opening a new file,
                # which is disallowed by resource.RLIMIT_NOFILE
                #
                # Of course, this isn't a foolproof solution by any means,
                # and it might lead to UNEXPECTED FAILURES later in execution.
                del sys.modules["os"]
                del sys.modules["os.path"]
                del sys.modules["sys"]

            self.run(script_str, user_globals, user_globals)
        # sys.exit ...
        except SystemExit:
            # sys.exit(0)
            raise bdb.BdbQuit
        except:
            if DEBUG:
                traceback.print_exc()

            trace_entry = dict(event="uncaught_exception")

            (exc_type, exc_val, exc_tb) = sys.exc_info()
            if hasattr(exc_val, "lineno"):
                trace_entry["line"] = exc_val.lineno
            if hasattr(exc_val, "offset"):
                trace_entry["offset"] = exc_val.offset

            trace_entry["exception_msg"] = type(exc_val).__name__ + ": " + str(exc_val)

            # SUPER SUBTLE! if ANY exception has already been recorded by
            # the program, then DON'T record it again as an uncaught_exception.
            # This looks kinda weird since the exact exception message doesn't
            # need to match up, but in practice, there should be at most only
            # ONE exception per trace.
            already_caught = False
            for e in self.trace:
                if e["event"] == "exception":
                    already_caught = True
                    break

            if not already_caught:
                if not self.done:
                    self.trace.append(trace_entry)

            raise bdb.BdbQuit  # need to forceably STOP execution

    def force_terminate(self):
        # self.finalize()
        raise bdb.BdbQuit  # need to forceably STOP execution

    def finalize(self):
        sys.stdout = self.GAE_STDOUT  # very important!
        sys.stderr = self.ORIGINAL_STDERR

        assert len(self.trace) <= (MAX_EXECUTED_LINES + 1)

        # don't do this anymore ...
        """
      # filter all entries after 'return' from '<module>', since they
      # seem extraneous:
      res = []
      for e in self.trace:
        res.append(e)
        if e['event'] == 'return' and e['func_name'] == '<module>':
          break
      """

        res = self.trace

        # if the SECOND to last entry is an 'exception'
        # and the last entry is return from <module>, then axe the last
        # entry, for aesthetic reasons :)
        if (
            len(res) >= 2
            and res[-2]["event"] == "exception"
            and res[-1]["event"] == "return"
            and res[-1]["func_name"] == "<module>"
        ):
            res.pop()

        self.trace = res

        return self.finalizer_func(self.executed_script, self.trace)


# the MAIN meaty function!!!

def exec_script_str(script_str, raw_input_lst_json, options_json, finalizer_func):
    options = json.loads(options_json)

    py_crazy_mode = "py_crazy_mode" in options and options["py_crazy_mode"]

    logger = PGLogger(
        options["cumulative_mode"],
        options["heap_primitives"],
        options["show_only_outputs"],
        finalizer_func,
        crazy_mode=py_crazy_mode,
    )

    # TODO: refactor these NOT to be globals
    global input_string_queue
    input_string_queue = []
    if raw_input_lst_json:
        # TODO: if we want to support unicode, remove str() cast
        input_string_queue = [str(e) for e in json.loads(raw_input_lst_json)]

    global __html__, __css__, __js__
    __html__, __css__, __js__ = None, None, None

    try:
        logger._runscript(script_str)
    except bdb.BdbQuit:
        pass
    finally:
        logger.finalize()


# disables security check and returns the result of finalizer_func
# WARNING: ONLY RUN THIS LOCALLY and never over the web, since
# security checks are disabled
def exec_script_str_local(
    script_str, raw_input_lst_json, cumulative_mode, heap_primitives, finalizer_func
):
    # TODO: add py_crazy_mode option here too ...
    logger = PGLogger(
        cumulative_mode,
        heap_primitives,
        False,
        finalizer_func,
        disable_security_checks=True,
    )

    # TODO: refactor these NOT to be globals
    global input_string_queue
    input_string_queue = []
    if raw_input_lst_json:
        # TODO: if we want to support unicode, remove str() cast
        input_string_queue = [str(e) for e in json.loads(raw_input_lst_json)]

    global __html__, __css__, __js__
    __html__, __css__, __js__ = None, None, None

    try:
        logger._runscript(script_str)
    except bdb.BdbQuit:
        pass
    finally:
        return logger.finalize()


def exec_str_with_user_ns(script_str, user_ns, finalizer_func):
    logger = PGLogger(False, False, False, finalizer_func, disable_security_checks=True)

    global __html__, __css__, __js__
    __html__, __css__, __js__ = None, None, None

    try:
        logger._runscript(script_str, user_ns)
    except bdb.BdbQuit:
        pass
    finally:
        return logger.finalize()



from enum import Enum

class LanguageOptions(Enum):
    PYTHON = ("python", "Python (in browser)")
    JAVA = ("java", "Java")
    CPP = ("cpp", "C++")
    C = ("c", "C")
    JAVASCRIPT = ("javascript", "Javascript")
    HTML = ("html", "HTML")
    SQL = ("sql", "SQL")

    def to_dict(self):
        return {
            "value": self.value[0],
            "label": self.value[1],
        }

    def value_only(self):
        return self.value[0]

    @classmethod
    def all_types(cls):
        return [member for member in cls]



# Copyright (C) 2013  Bradley N. Miller, Barbara Ericson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import pdb

__author__ = "bmiller"

from docutils import nodes
from docutils.parsers.rst import directives
from .assessbase import Assessment
from runestone.common.runestonedirective import (
    RunestoneNode,
    RunestoneIdNode,
    get_node_line,
)
from runestone.server.componentdb import (
    addQuestionToDB,
    addHTMLToDB,
    maybeAddToAssignment,
)


class MChoiceNode(nodes.General, nodes.Element, RunestoneIdNode):
    pass


XML_START = """
    <exercise label="{divid}">
        <statement>

"""

XML_START_END = """
        </statement>
"""

XML_OPTION = """
            <choice {is_correct}>
                <statement>
                    <p>{atext}</p>
                </statement>
                <feedback>
                    <p>{feedtext}</p>
                </feedback>
            </choice>
"""

XML_END = """
    </exercise>
"""

TEMPLATE_START = """
    <div class="{divclass} {optclass}">
    <ul data-component="multiplechoice" data-question_label="{question_label}" data-multipleanswers="{multipleAnswers}" {random} id="{divid}" {optional} style="visibility: hidden;">
    """

OPTION = """
    <li data-component="answer" {is_correct} id="{divid}_opt_{alabel}">{atext}</li><li data-component="feedback">{feedtext}</li>
    """

TEMPLATE_END = """

    </ul>
    </div>
    """


def visit_mc_common(self, node):

    res = ""
    if "random" in node["runestone_options"]:
        node["runestone_options"]["random"] = "data-random"
    else:
        node["runestone_options"]["random"] = ""
    # Use multiple_answers behavior if explicitly required or if multiple correct answers were provided.
    if ("multiple_answers" in node["runestone_options"]) or (
        "," in node["runestone_options"]["correct"]
    ):
        node["runestone_options"]["multipleAnswers"] = "true"
    else:
        node["runestone_options"]["multipleAnswers"] = "false"
    res = node["template_start"].format(**node["runestone_options"])

    return res


def depart_mc_common(self, node):
    res = ""
    currFeedback = ""
    # Add all of the possible answers
    if node["runestone_options"]["is_new_style"] == False and hasattr(self, "output"):
        res += "<choices>\n"
    okeys = list(node["runestone_options"].keys())
    okeys.sort()
    for k in okeys:
        if "answer_" in k:
            x, label = k.split("_")
            node["runestone_options"]["alabel"] = label
            node["runestone_options"]["atext"] = node["runestone_options"][k]
            currFeedback = "feedback_" + label
            node["runestone_options"]["feedtext"] = node["runestone_options"].get(
                currFeedback, ""
            )  # node["runestone_options"][currFeedback]
            if label in node["runestone_options"]["correct"]:
                node["runestone_options"]["is_correct"] = "data-correct='yes'"
            else:
                node["runestone_options"]["is_correct"] = ""
            res += node["template_option"].format(**node["runestone_options"])
    if node["runestone_options"]["is_new_style"] == False and hasattr(self, "output"):
        res += "</choices>\n"
    res += node["template_end"].format(**node["runestone_options"])
    return res


def visit_mc_xml(self, node):
    node["template_start"] = XML_START
    node["template_end"] = XML_END
    node["template_option"] = XML_OPTION
    res = visit_mc_common(self, node)
    self.output.append(res)


def depart_mc_xml(self, node):
    if not node["runestone_options"]["is_new_style"]:
        self.output.append(XML_START_END)
    res = depart_mc_common(self, node)
    res = res.replace("data-correct", "correct")
    self.output.append(res)


def visit_mc_html(self, node):
    node["delimiter"] = "_start__{}_".format(node["runestone_options"]["divid"])
    self.body.append(node["delimiter"])
    res = visit_mc_common(self, node)
    self.body.append(res)


def depart_mc_html(self, node):
    res = depart_mc_common(self, node)
    self.body.append(res)

    addHTMLToDB(
        node["runestone_options"]["divid"],
        node["runestone_options"]["basecourse"],
        "".join(self.body[self.body.index(node["delimiter"]) + 1 :]),
    )

    self.body.remove(node["delimiter"])


#####################
# multiple choice question with multiple feedback
# author - Barb Ericson
# author - Anusha
class MChoice(Assessment):
    r"""
    The syntax for a multiple-choice question is:

    .. mchoice:: uniqueid
        :multiple_answers: [optional]. Implied if ``:correct:`` contains a list.
        :random: [optional]

        The following arguments supply answers and feedback. See below for an alternative method of specification.

        :correct: letter of correct answer or list of correct answer letters (in case of multiple answers)
        :answer_a: possible answer  -- what follows the _ is the answer's label.
        :answer_b: possible answer
        :answer_c: possible answer
        :answer_d: possible answer
        :answer_e: possible answer
        :feedback_a: displayed if a is picked
        :feedback_b: displayed if b is picked
        :feedback_c: displayed if c is picked
        :feedback_d: displayed if d is picked
        :feedback_e: displayed if e is picked

        Question text; this may contain multiple paragraphs with any markup.

        An alternative method of specifying answers and feedback: Place an `unordered list <http://www.sphinx-doc.org/en/stable/rest.html#lists-and-quote-like-blocks>`_ at the end of the question text, in the following format. Note: If your question text happens to end with an unordered list, then place a comment, consisting of a paragraph containing only ``..`` at the end of the list. For example:

        -   This list is still part of the question text.

        ..

        -   Text for answer A.

            Your text may be multiple paragraphs, including `images <http://www.sphinx-doc.org/en/stable/rest.html#images>`_
            and any other `inline <http://www.sphinx-doc.org/en/stable/rest.html#inline-markup>`_ or block markup. For example: :math:`\sqrt(2)/2`. As earlier, if your feedback contains an unordered list, end it with a comment.

            -   For example, this is part of the answer text.

            ..

            +   This is feedback for answer A. This is a correct answer because the bullet is a ``+``.

                This may also span multiple paragraphs and include any markup.
                However, there can be only one item in this unordered list.

        -   Text for answer B.

            -   Feedback for answer B. This is an incorrect answer, because the bullet is not a ``+``.
        -   Text for answer C. Note that the empty line between a sublist and a list may be omitted.

            +   Feedback for answer C, which is a correct answer. However, the empty line is required between a list and a sublist.

        -   ... and so on.

            -   Up to 26 answers and feedback pairs may be provided.

    config values (conf.py):

    - mchoice_div_class - custom CSS class of the component's outermost div
    """

    required_arguments = 1
    optional_arguments = 1
    final_argument_whitespace = True
    has_content = True
    option_spec = Assessment.option_spec.copy()
    option_spec.update(
        {
            "answer_a": directives.unchanged,
            "answer_b": directives.unchanged,
            "answer_c": directives.unchanged,
            "answer_d": directives.unchanged,
            "answer_e": directives.unchanged,
            "correct": directives.unchanged,
            "feedback_a": directives.unchanged,
            "feedback_b": directives.unchanged,
            "feedback_c": directives.unchanged,
            "feedback_d": directives.unchanged,
            "feedback_e": directives.unchanged,
            "random": directives.flag,
            "multiple_answers": directives.flag,
        }
    )

    def run(self):
        """
        process the multiplechoice directive and generate html for output.

        :param self:
        :return: An MChoiceNode.
        """

        super(MChoice, self).run()

        addQuestionToDB(self)

        mcNode = MChoiceNode()
        mcNode["runestone_options"] = self.options
        mcNode["source"], mcNode["line"] = self.state_machine.get_source_and_line(
            self.lineno
        )
        mcNode["template_start"] = TEMPLATE_START
        mcNode["template_option"] = OPTION
        mcNode["template_end"] = TEMPLATE_END
        # For MChoice its better to insert the qnum into the content before further processing.
        self.updateContent()

        self.state.nested_parse(self.content, self.content_offset, mcNode)
        env = self.state.document.settings.env
        self.options["divclass"] = env.config.mchoice_div_class
        # Expected _`structure`, with assigned variable names and transformations made:
        #
        # .. code-block::
        #   :number-lines:
        #
        #   mcNode = MChoiceNode()
        #       Item 1 of problem text
        #       ...
        #       Item n of problem text
        #       answers_bullet_list = bullet_list() -> AnswersBulletList() <-- last item of mcNode may be a bulleted list of answers and feedback.
        #           answer_list_item = list_item() -> AnswerListItem()
        #               Item 1 of text for answer A
        #               ...
        #               Item n of text for answer A
        #               feedback_bullet_list = bullet_list() -> FeedbackBulletList() <-- last item must be a bulleted list containing feedback.
        #                   feedback_list_item = list_item() -> FeedbackListItem()   <-- A single item in the list, which contains the feedback.
        #           answer_list_item = list_item() -> AnswerListItem()
        #               Item 1 of text for answer B
        #               ...and so on...
        #
        # See if the last item is a list. If so, and questions/answers weren't specified as options, assume it contains questions and answers.
        self.options["is_new_style"] = False
        answers_bullet_list = mcNode[-1] if len(mcNode) else None
        if isinstance(answers_bullet_list, nodes.bullet_list) and (
            "answer_a" not in self.options and ("correct" not in self.options)
        ):
            # Accumulate the correct answers.
            correct_answers = []
            self.options["is_new_style"] = True
            # Walk it, processing each answer and its associated feedback.
            for answer_list_item in answers_bullet_list:
                assert isinstance(answer_list_item, nodes.list_item)

                # Look for the feedback for this answer -- the last child of this answer list item.
                feedback_bullet_list = answer_list_item[-1]
                if (
                    not isinstance(feedback_bullet_list, nodes.bullet_list)
                    # It should have just one item (the feedback itself).
                    or (len(feedback_bullet_list) != 1)
                ):
                    raise self.error(
                        "On line {}, a single-item list must be nested under each answer.".format(
                            get_node_line(feedback_bullet_list)
                        )
                    )

                # Check for a correct answer.
                feedback_list_item = feedback_bullet_list[0]
                assert isinstance(feedback_list_item, nodes.list_item)
                if feedback_bullet_list["bullet"] == "+":
                    correct_answers.append(
                        chr(answer_list_item.parent.index(answer_list_item) + ord("a"))
                    )

                # Change the feedback list item (which is currently a generic list_item) to our special node class (a FeedbackListItem).
                feedback_list_item.replace_self(
                    FeedbackListItem(
                        feedback_list_item.rawsource,
                        *feedback_list_item.children,
                        **feedback_list_item.attributes
                    )
                )

                # Change the feedback bulleted list (currently a bullet_list) to our class (a FeedbackBulletList).
                feedback_bullet_list.replace_self(
                    FeedbackBulletList(
                        feedback_bullet_list.rawsource,
                        *feedback_bullet_list.children,
                        **feedback_bullet_list.attributes
                    )
                )

                # Change the answer list item (currently a list_item) to an AnswerListItem.
                answer_list_item.replace_self(
                    AnswerListItem(
                        answer_list_item.rawsource,
                        *answer_list_item.children,
                        **answer_list_item.attributes
                    )
                )

            # Change the answer bulleted list (currently a bullet_list) to an AnswersBulletList.
            answers_bullet_list.replace_self(
                AnswersBulletList(
                    answers_bullet_list.rawsource,
                    *answers_bullet_list.children,
                    **answers_bullet_list.attributes
                )
            )
            # Store the correct answers.
            self.options["correct"] = ",".join(correct_answers)

        maybeAddToAssignment(self)
        # Check that a correct answer was provided.
        if not self.options.get("correct"):
            raise self.error("No correct answer specified.")
        return [mcNode]


# Define a bullet_list which contains answers (see the structure_).
class AnswersBulletList(nodes.bullet_list, RunestoneNode):
    pass


# Define a list_item which contains answers (see the structure_).
class AnswerListItem(nodes.list_item, RunestoneNode):
    pass


# Define a bullet_list which contains feedback (see the structure_).
class FeedbackBulletList(nodes.bullet_list, RunestoneNode):
    pass


# Define a list_item which contains answers (see the structure_).
class FeedbackListItem(nodes.list_item, RunestoneNode):
    pass


# The ``<ul>`` tag will be generated already -- don't output it.
def visit_answers_bullet_html(self, node):
    # Prevent the list items, which are wrapped in ``<paragraph>`` tags, from emitting the ``<p>``. See similar code in ``docutils.writers._html_base.HTMLTranslator.visit_bullet_list`` and its use in ``docutils.writer.html4css1.HTMLTranslator.visit_paragraph``.
    self.context.append((self.compact_simple, self.compact_p))
    self.compact_p = None
    self.compact_simple = True


# The ``</ul>`` tag will be generated already -- don't output it.
def depart_answers_bullet_html(self, node):
    # Restore the state modified in ``visit_answers_bullet_html``.
    self.compact_simple, self.compact_p = self.context.pop()


def visit_answers_bullet_xml(self, node):
    self.output.append("</statement>")
    self.output.append("<choices>")


def depart_answers_bullet_xml(self, node):
    self.output.append("</choices>")


# Write out the special attributes needed by the ``<li>`` tag.
def visit_answer_list_item(self, node):
    # See the structure_.
    mc_node = node.parent.parent

    # _`label`: Turn the index of this item in the answer_bullet_list (see structure_) into a letter.
    label = chr(node.parent.index(node) + ord("a"))
    # Update dict for formatting the HTML.
    mc_node["runestone_options"]["alabel"] = label
    if label in mc_node["runestone_options"]["correct"]:
        mc_node["runestone_options"]["is_correct"] = "data-correct"
    else:
        mc_node["runestone_options"]["is_correct"] = ""

    # Format the HTML.
    self.body.append(
        '<li data-component="answer" {is_correct} id="{divid}_opt_{alabel}">'.format(
            **mc_node["runestone_options"]
        )
    )


def visit_answer_list_item_xml(self, node):
    # See the structure_.
    mc_node = node.parent.parent

    # _`label`: Turn the index of this item in the answer_bullet_list (see structure_) into a letter.
    label = chr(node.parent.index(node) + ord("a"))
    # Update dict for formatting the HTML.
    mc_node["runestone_options"]["alabel"] = label
    if label in mc_node["runestone_options"]["correct"]:
        mc_node["runestone_options"]["is_correct"] = "correct"
        self.output.append("<choice correct='yes'>")
    else:
        mc_node["runestone_options"]["is_correct"] = ""
        self.output.append("<choice>")
    # Format the HTML.
    self.output.append(
        '<statement id="{divid}_opt_{alabel}">'.format(**mc_node["runestone_options"])
    )


# Although the feedback for an answer is given as a sublist, the HTML is just a list. So, let the feedback list item close this list.
def depart_answer_list_item(self, node):
    pass


def depart_answer_list_item_xml(self, node):
    self.output.append("</choice>")


# Nothing to output, since feedback isn't nested under an answer in the HTML.
def visit_feedback_bullet_html(self, node):
    pass


# Nothing to output, since feedback isn't nested under an answer in the HTML.
def depart_feedback_bullet_html(self, node):
    pass


def visit_feedback_list_item(self, node):
    # See label_ and structure_.
    answer_list_item = node.parent.parent
    mc_node = answer_list_item.parent.parent
    label = chr(answer_list_item.parent.index(answer_list_item) + ord("a"))
    mc_node["runestone_options"]["alabel"] = label
    self.body.append(
        '</li><li data-component="feedback" id="{divid}_opt_{alabel}">\n'.format(
            **mc_node["runestone_options"]
        )
    )


def visit_feedback_list_item_xml(self, node):
    # See label_ and structure_.
    answer_list_item = node.parent.parent
    mc_node = answer_list_item.parent.parent
    label = chr(answer_list_item.parent.index(answer_list_item) + ord("a"))
    mc_node["runestone_options"]["alabel"] = label
    self.output.append("</statement>")
    self.output.append(
        '<feedback id="{divid}_opt_{alabel}">\n'.format(**mc_node["runestone_options"])
    )


def depart_feedback_list_item(self, node):
    self.body.append("</li>")


def depart_feedback_list_item_xml(self, node):
    self.output.append("</feedback>")


# backwards compatibility


class MChoiceMF(MChoice):
    def run(self):
        raise self.error(
            "This directive has been depreciated. Please convert to the new directive 'mchoice'"
        )
        mcmfNode = super(MChoiceMF, self).run()[0]

        return [mcmfNode]


class MChoiceMA(MChoice):
    def run(self):
        self.options["multiple_answers"] = "multipleAnswers"
        raise self.error(
            "This directive has been depreciated. Please convert to the new directive 'mchoice'"
        )
        mchoicemaNode = super(MChoiceMA, self).run()[0]

        return [mchoicemaNode]


class MChoiceRandomMF(MChoice):
    def run(self):
        self.options["random"] = "random"
        raise self.error(
            "This directive has been depreciated. Please convert to the new directive 'mchoice'"
        )
        mchoicerandommfNode = super(MChoiceRandomMF, self).run()[0]

        return [mchoicerandommfNode]



# Copyright (C) 2011  Bradley N. Miller
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__author__ = "bmiller"

from docutils import nodes
from docutils.parsers.rst import directives
from runestone.common.runestonedirective import RunestoneDirective


def setup(app):
    app.add_directive("shortname", Meta)
    app.add_directive("description", Meta)


class Meta(RunestoneDirective):
    required_arguments = 1
    optional_arguments = 50

    def run(self):
        """
        process the video directive and generate html for output.
        :param self:
        :return:
        """
        raw_node = nodes.raw(self.block_text, "", format="html")
        raw_node.source, raw_node.line = self.state_machine.get_source_and_line(
            self.lineno
        )
        return [raw_node]


source = """\
This is some text.

.. shortname:: divid
.. description::  foo bar baz

This is some more text.
"""

if __name__ == "__main__":
    from docutils.core import publish_parts

    directives.register_directive("shortname", Meta)
    directives.register_directive("description", Meta)

    doc_parts = publish_parts(
        source,
        settings_overrides={"output_encoding": "utf8", "initial_header_level": 2},
        writer_name="html",
    )

    print(doc_parts["html_body"])



# Copyright (C) 2011  Bradley N. Miller
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__author__ = "bmiller"

from docutils import nodes
from docutils.parsers.rst import directives
from runestone.common.runestonedirective import (
    RunestoneDirective,
    RunestoneIdDirective,
)
from .multiplechoice import *


def setup(app):
    app.add_directive("mchoice", MChoice)
    app.add_directive("mchoicemf", MChoiceMF)
    app.add_directive("mchoicema", MChoiceMA)
    app.add_directive("mchoicerandommf", MChoiceRandomMF)
    app.add_directive("addbutton", AddButton)
    app.add_directive("qnum", QuestionNumber)

    app.add_config_value("mchoice_div_class", "runestone", "html")

    app.add_node(
        MChoiceNode,
        html=(visit_mc_html, depart_mc_html),
        xml=(visit_mc_xml, depart_mc_xml),
    )

    app.add_node(
        AnswersBulletList,
        html=(visit_answers_bullet_html, depart_answers_bullet_html),
        xml=(visit_answers_bullet_xml, depart_answers_bullet_xml),
    )
    app.add_node(
        AnswerListItem,
        html=(visit_answer_list_item, depart_answer_list_item),
        xml=(visit_answer_list_item_xml, depart_answer_list_item_xml),
    )
    app.add_node(
        FeedbackBulletList,
        html=(visit_feedback_bullet_html, depart_feedback_bullet_html),
        xml=(visit_feedback_bullet_html, depart_feedback_bullet_html),
    )
    app.add_node(
        FeedbackListItem,
        html=(visit_feedback_list_item, depart_feedback_list_item),
        xml=(visit_feedback_list_item_xml, depart_feedback_list_item_xml),
    )


class AddButton(RunestoneIdDirective):
    required_arguments = 1
    optional_arguments = 1
    final_argument_whitespace = True
    has_content = True

    def run(self):
        """
        :param self:
        :return:
        .. addbutton:: bname

        ...
        """
        super(AddButton, self).run()

        TEMPLATE_START = """
            <div id="%(divid)s" class="alert alert-warning %(optclass)s">
            <form name="%(divid)s_form" method="get" action="" onsubmit="return false;">
            """

        TEMPLATE_END = """
            <button class='btn btn-inverse' name="reset" onclick="resetPage('%(divid)s')">Forget My Answers</button>
            </form>
            </div>
            """

        res = ""
        res = TEMPLATE_START % self.options

        res += TEMPLATE_END % self.options
        rawnode = nodes.raw(self.block_text, res, format="html")
        rawnode.source, rawnode.line = self.state_machine.get_source_and_line(
            self.lineno
        )
        return [rawnode]


class QuestionNumber(RunestoneDirective):
    """
    Set Parameters for Question Numbering:

    .. code-block:: rst
        :linenos:

        .. qnum::
            'prefix': character prefix before the number
            'suffix': character prefix after the number
            'start': start numbering with this value

    For example:

    .. code-block:: rst
        :linenos:

        .. qnum::
            :prefix: turtle-
            :start: 10
    """

    required_arguments = 0
    optional_arguments = 3
    has_content = False
    option_spec = RunestoneDirective.option_spec.copy()
    option_spec.update(
        {
            "prefix": directives.unchanged,
            "suffix": directives.unchanged,
            "start": directives.positive_int,
        }
    )

    def run(self):
        env = self.state.document.settings.env

        if "start" in self.options:
            env.assesscounter = self.options["start"] - 1

        if "prefix" in self.options:
            env.assessprefix = self.options["prefix"]

        if "suffix" in self.options:
            env.assesssuffix = self.options["suffix"]

        return []
